import json
import os
import pickle
import time
import unittest.mock
import uuid
from typing import Any, Dict, cast
from unittest.mock import Mock, patch

import boto3
import pytest
from freezegun import freeze_time
from moto import mock_logs

import redun.executors.aws_batch
import redun.tests.utils
from redun import File, job_array, task
from redun.cli import RedunClient, import_script
from redun.config import Config
from redun.executors.aws_batch import (
    BATCH_JOB_STATUSES,
    BATCH_LOG_GROUP,
    FAILED,
    SUCCEEDED,
    AWSBatchExecutor,
    batch_submit,
    equiv_job_def,
    get_batch_job_name,
    get_hash_from_job_name,
    get_job_definition,
    get_job_details,
    get_job_log_stream,
    get_or_create_job_definition,
    iter_batch_job_log_lines,
    iter_batch_job_logs,
    make_job_def_name,
    parse_job_error,
    submit_task,
)
from redun.executors.code_packaging import create_tar, package_code
from redun.executors.command import REDUN_REQUIRED_VERSION
from redun.executors.scratch import (
    SCRATCH_ERROR,
    SCRATCH_HASHES,
    SCRATCH_INPUT,
    SCRATCH_OUTPUT,
    ExceptionNotFoundError,
    get_array_scratch_file,
    get_job_scratch_file,
)
from redun.file import Dir
from redun.scheduler import Execution, Job, Scheduler, Traceback
from redun.scripting import ScriptError
from redun.tests.utils import mock_s3, mock_scheduler, use_tempdir, wait_until
from redun.utils import pickle_dumps


def test_make_job_def_name() -> None:
    """
    Job definition names should autogenerate from docker image names.
    """
    assert make_job_def_name("my-image") == "my-image-jd"
    assert make_job_def_name("my-image", "-job-def") == "my-image-job-def"
    assert make_job_def_name("my.image") == "myimage-jd"
    assert make_job_def_name("a" * 200) == ("a" * 125) + "-jd"


@patch("redun.executors.aws_utils.get_aws_client")
def test_get_job_definition(get_aws_client_mock) -> None:
    """
    Most recent active revision should be returned or empty dict if no matching job defs.
    """
    # Simulate case where no matching jobs are returned from batch API.
    get_aws_client_mock.return_value.describe_job_definitions.return_value = {"jobDefinitions": []}
    assert get_job_definition("JOB_DEF_1") == {}

    # Simulate case where there are multiple revisions, confirm newest is returned.
    get_aws_client_mock.return_value.describe_job_definitions.return_value = {
        "jobDefinitions": [
            {"revision": 1, "jobDefinitionArn": "ARN1"},
            {"revision": 3, "jobDefinitionArn": "ARN3"},
            {"revision": 2, "jobDefinitionArn": "ARN2"},
        ]
    }
    # Need to change the job def name here from the first case above to avoid the lru cache on
    # get_job_definition.
    job_def = get_job_definition("JOB_DEF_2")
    assert job_def["jobDefinitionArn"] == "ARN3"


@patch("redun.executors.aws_utils.get_aws_client")
@patch("redun.executors.aws_batch.get_job_definition")
def test_required_job_def_name(get_job_definition_mock, _) -> None:
    """
    Confirm that job_def_name is required when autocreate_job_def is False.
    """
    # A job_def_name is required when autocreate is False.
    with pytest.raises(AssertionError):
        batch_submit({"command": ["ls"]}, "queue", "image", autocreate_job_def=False)

    # When the required job_def_name is supplied, an error should be raised if a matching
    # definition cannot be found.
    get_job_definition_mock.return_value = {}
    with pytest.raises(ValueError):
        batch_submit(
            {"command": ["ls"]},
            "queue",
            "image",
            job_def_name="NONEXISTENT",
            autocreate_job_def=False,
        )


@pytest.mark.parametrize("array,suffix", [(False, ""), (True, "-array")])
def test_get_hash_from_job_name(array, suffix) -> None:
    """
    Returns the Job hash from a AWS Batch job name.
    """
    prefix = "my-job-prefix"
    job_hash = "c000d7f9b6275c58aff9d5466f6a1174e99195ca"
    job_name = get_batch_job_name(prefix, job_hash, array=array)
    assert job_name.startswith(prefix)
    assert job_name.endswith(suffix)

    job_hash2 = get_hash_from_job_name(job_name)
    assert job_hash2 == job_hash


@patch("redun.executors.aws_utils.get_aws_client")
def test_get_or_create_job_definition(get_aws_client_mock) -> None:
    """
    Confirm that jobs are correctly fetched or created.
    """

    # Get mocks.
    describe_job_definitions = get_aws_client_mock.return_value.describe_job_definitions
    register_job_definition = get_aws_client_mock.return_value.register_job_definition

    # Let's skip the LRU cache for this test.
    _get_or_create_job_definition = get_or_create_job_definition.__wrapped__

    # 1. Jobs get created
    describe_job_definitions.return_value = {"jobDefinitions": []}
    register_job_definition.return_value = {"jobId": "new-job"}
    result = _get_or_create_job_definition(
        job_def_name="test-jd", image="an-image:latest", role="aRole"
    )

    assert result == {"jobId": "new-job"}
    job_def = dict(
        jobDefinitionName="test-jd",
        type="container",
        containerProperties={
            "command": ["ls"],
            "image": "an-image:latest",
            "jobRoleArn": "aRole",
            "environment": [],
            "mountPoints": [],
            "volumes": [],
            "resourceRequirements": [
                {"type": "VCPU", "value": "1"},
                {"type": "MEMORY", "value": "4096"},
            ],
            "ulimits": [],
            "privileged": False,
        },
    )
    register_job_definition.assert_called_with(**job_def)

    # 2a. Identical jobs are not recreated
    describe_job_definitions.return_value = {"jobDefinitions": [job_def]}
    register_job_definition.reset_mock()
    _get_or_create_job_definition(job_def_name="test-jd", image="an-image:latest", role="aRole")
    register_job_definition.assert_not_called()

    # 2.b. resource differences are ignored on single node
    _get_or_create_job_definition(
        job_def_name="test-jd",
        image="an-image:latest",
        role="aRole",
        vcpus=2,
        memory=8,
    )
    register_job_definition.assert_not_called()

    # 2.c. The newer revision is preferred.
    describe_job_definitions.return_value = {
        "jobDefinitions": [
            {**job_def, "jobDefinitionName": "test-jd:3"},
            {**job_def, "jobDefinitionName": "test-jd:10"},
            {**job_def, "jobDefinitionName": "test-jd:1"},
            {**job_def, "jobDefinitionName": "test-jd:2"},
        ]
    }
    register_job_definition.reset_mock()
    result = _get_or_create_job_definition(
        job_def_name="test-jd",
        image="an-image:latest",
        role="aRole",
    )
    assert result["jobDefinitionName"] == "test-jd:10"
    register_job_definition.assert_not_called()

    # 3. Extra keys are ignored
    describe_job_definitions.return_value = {
        "jobDefinitions": [{"extraKey": "ignored", **job_def}]
    }
    register_job_definition.reset_mock()
    _get_or_create_job_definition(job_def_name="test-jd3", image="an-image:latest", role="aRole")
    register_job_definition.assert_not_called()

    # 4. Important differences trigger recreation
    # 4.a. job type
    describe_job_definitions.return_value = {
        "jobDefinitions": [
            {
                "jobDefinitionName": "test-jd4a",
                "type": "other",
                "containerProperties": {
                    "command": ["ls"],
                    "image": "an-image:latest",
                    "jobRoleArn": "aRole",
                    "environment": [],
                    "mountPoints": [],
                    "volumes": [],
                    "resourceRequirements": [
                        {"type": "VCPU", "value": "2"},
                        {"type": "MEMORY", "value": "4096"},
                    ],
                    "ulimits": [],
                    "privileged": False,
                },
            }
        ]
    }
    register_job_definition.reset_mock()
    result = _get_or_create_job_definition(
        job_def_name="test-jd4a", image="an-image:latest", role="aRole"
    )
    assert result == {"jobId": "new-job"}
    register_job_definition.assert_called_with(
        jobDefinitionName="test-jd4a",
        type="container",
        containerProperties={
            "command": ["ls"],
            "image": "an-image:latest",
            "jobRoleArn": "aRole",
            "environment": [],
            "mountPoints": [],
            "volumes": [],
            "resourceRequirements": [
                {"type": "VCPU", "value": "1"},
                {"type": "MEMORY", "value": "4096"},
            ],
            "ulimits": [],
            "privileged": False,
        },
    )

    # 4.b. num_nodes
    describe_job_definitions.return_value = {
        "jobDefinitions": [
            {
                "jobDefinitionName": "test-jd4b",
                "type": "container",
                "containerProperties": {
                    "command": ["ls"],
                    "image": "an-image:latest",
                    "jobRoleArn": "aRole",
                    "environment": [],
                    "mountPoints": [],
                    "volumes": [],
                    "resourceRequirements": [
                        {"type": "VCPU", "value": "1"},
                        {"type": "MEMORY", "value": "4096"},
                    ],
                    "ulimits": [],
                    "privileged": False,
                },
            }
        ]
    }
    register_job_definition.reset_mock()
    _get_or_create_job_definition(
        job_def_name="test-jd4b", image="an-image:latest", role="aRole", num_nodes=3
    )
    multinode_job = {
        "jobDefinitionName": "test-jd4b",
        "type": "multinode",
        "nodeProperties": {
            "mainNode": 0,
            "numNodes": 3,
            "nodeRangeProperties": [
                {
                    "container": {
                        "command": ["ls"],
                        "image": "an-image:latest",
                        "jobRoleArn": "aRole",
                        "environment": [],
                        "mountPoints": [],
                        "volumes": [],
                        "resourceRequirements": [
                            {"type": "VCPU", "value": "1"},
                            {"type": "MEMORY", "value": "4096"},
                        ],
                        "ulimits": [
                            {
                                "name": "nofile",
                                "softLimit": 65535,
                                "hardLimit": 65535,
                            }
                        ],
                        "privileged": False,
                    },
                    "targetNodes": "0",
                },
                {
                    "container": {
                        "command": ["ls"],
                        "image": "an-image:latest",
                        "jobRoleArn": "aRole",
                        "environment": [],
                        "mountPoints": [],
                        "volumes": [],
                        "resourceRequirements": [
                            {"type": "VCPU", "value": "1"},
                            {"type": "MEMORY", "value": "4096"},
                        ],
                        "ulimits": [
                            {
                                "name": "nofile",
                                "softLimit": 65535,
                                "hardLimit": 65535,
                            }
                        ],
                        "privileged": False,
                    },
                    "targetNodes": "1:",
                },
            ],
        },
    }
    register_job_definition.assert_called_with(**multinode_job)

    # 4c. Multinode resource definitions do not trigger changes
    describe_job_definitions.return_value = {"jobDefinitions": [multinode_job]}
    register_job_definition.reset_mock()
    _get_or_create_job_definition(
        job_def_name="test-jd4b",
        image="an-image:latest",
        role="aRole",
        num_nodes=3,
        vcpus=2,
        memory=8,
    )
    register_job_definition.assert_not_called()

    # 4.e. image
    describe_job_definitions.return_value = {
        "jobDefinitions": [
            {
                "jobDefinitionName": "test-jd4a",
                "type": "other",
                "containerProperties": {
                    "command": ["ls"],
                    "image": "an-image:latest",
                    "jobRoleArn": "aRole",
                    "environment": [],
                    "mountPoints": [],
                    "volumes": [],
                    "resourceRequirements": [
                        {"type": "VCPU", "value": "2"},
                        {"type": "MEMORY", "value": "4096"},
                    ],
                    "ulimits": [],
                    "privileged": False,
                },
            }
        ]
    }
    register_job_definition.reset_mock()
    result = _get_or_create_job_definition(
        job_def_name="test-jd4a", image="an-image:custom", role="aRole"
    )
    assert result == {"jobId": "new-job"}
    register_job_definition.assert_called_with(
        jobDefinitionName="test-jd4a",
        type="container",
        containerProperties={
            "command": ["ls"],
            "image": "an-image:custom",
            "jobRoleArn": "aRole",
            "environment": [],
            "mountPoints": [],
            "volumes": [],
            "resourceRequirements": [
                {"type": "VCPU", "value": "1"},
                {"type": "MEMORY", "value": "4096"},
            ],
            "ulimits": [],
            "privileged": False,
        },
    )

    # 5. Passing extra job configuration options
    ulimit_option = {
        "containerProperties": {
            "ulimits": [
                {
                    "name": "nofile",
                    "softLimit": 2048,
                    "hardLimit": 2048,
                }
            ]
        }
    }

    describe_job_definitions.return_value = {"jobDefinitions": [job_def]}
    register_job_definition.reset_mock()
    _get_or_create_job_definition(
        job_def_name="test-jd",
        image="an-image:latest",
        role="aRole",
        job_def_extra=ulimit_option,
    )
    register_job_definition.assert_called_with(
        jobDefinitionName="test-jd",
        type="container",
        containerProperties={
            "command": ["ls"],
            "image": "an-image:latest",
            "jobRoleArn": "aRole",
            "environment": [],
            "mountPoints": [],
            "volumes": [],
            "resourceRequirements": [
                {"type": "VCPU", "value": "1"},
                {"type": "MEMORY", "value": "4096"},
            ],
            "ulimits": [
                {
                    "name": "nofile",
                    "softLimit": 2048,
                    "hardLimit": 2048,
                }
            ],
            "privileged": False,
        },
    )


@patch("redun.executors.aws_utils.get_aws_client")
def test_job_definition_role(get_aws_client_mock) -> None:
    """
    Confirm that jobs roles are set correctly.
    """

    # The default role is the AWS account's ecsTaskExecutionRole.
    details = get_job_details("image")
    assert details["containerProperties"]["jobRoleArn"].endswith("role/ecsTaskExecutionRole")

    # If specified, use the user's role.
    details = get_job_details("image", role="aRole")
    assert details["containerProperties"]["jobRoleArn"] == "aRole"

    # If the "none" role is specified, do not set the role.
    details = get_job_details("image", role="none")
    assert "jobRoleArn" not in details["containerProperties"]


@patch("redun.executors.aws_utils.get_aws_client")
@patch("redun.executors.aws_batch.get_or_create_job_definition")
def test_job_submit_resources(get_or_create_job_definition_mock, get_aws_client_mock) -> None:
    """
    Confirm that resource overrides are properly applied.
    """

    get_or_create_job_definition_mock.return_value = {"jobDefinitionArn": "test-jd-arn"}

    # 1. Single node
    batch_submit(
        batch_job_args={
            "containerOverrides": {"command": ["test_command"]},
            "user": "user-test",
        },
        job_name="test",
        queue="test_queue",
        image="test_image",
        vcpus=4,
        memory=8,
        gpus=2,
    )

    # Check that the resources are applied but the custom commands are specifically not touched.
    get_aws_client_mock.return_value.submit_job.assert_called_with(
        jobName="test",
        jobQueue="test_queue",
        jobDefinition="test-jd-arn",
        retryStrategy={"attempts": 1},
        containerOverrides={
            "command": ["test_command"],
            "resourceRequirements": [
                {"type": "VCPU", "value": "4"},
                {"type": "MEMORY", "value": "8192"},
                {"type": "GPU", "value": "2"},
            ],
        },
        propagateTags=True,
        user="user-test",
    )

    # 2. multi-node
    get_aws_client_mock.return_value.submit_job.reset_mock()
    batch_submit(
        batch_job_args={
            "nodeOverrides": {
                "nodePropertyOverrides": [
                    {
                        "targetNodes": "0",
                        "containerOverrides": {"command": ["command_rank_0"]},
                    },
                    {
                        "targetNodes": "1:",
                        "containerOverrides": {"command": ["command_rank_1"]},
                    },
                ],
            }
        },
        job_name="test",
        queue="test_queue",
        image="test_image",
        vcpus=4,
        memory=8,
        gpus=2,
    )

    # Check that the resources are applied but the custom commands are specifically not touched.
    get_aws_client_mock.return_value.submit_job.assert_called_with(
        jobName="test",
        jobQueue="test_queue",
        jobDefinition="test-jd-arn",
        retryStrategy={"attempts": 1},
        nodeOverrides={
            "nodePropertyOverrides": [
                {
                    "targetNodes": "0",
                    "containerOverrides": {
                        "command": ["command_rank_0"],
                        "resourceRequirements": [
                            {"type": "VCPU", "value": "4"},
                            {"type": "MEMORY", "value": "8192"},
                            {"type": "GPU", "value": "2"},
                        ],
                    },
                },
                {
                    "targetNodes": "1:",
                    "containerOverrides": {
                        "command": ["command_rank_1"],
                        "resourceRequirements": [
                            {"type": "VCPU", "value": "4"},
                            {"type": "MEMORY", "value": "8192"},
                            {"type": "GPU", "value": "2"},
                        ],
                    },
                },
            ]
        },
        propagateTags=True,
    )


def test_batch_tags(scheduler: Scheduler) -> None:
    """
    Executor should be able to determine batch tags for a batch job.
    """
    # Setup executor.
    config = Config(
        {
            "batch": {
                "image": "image",
                "queue": "queue",
                "s3_scratch": "s3_scratch_prefix",
                "batch_tags": '{"team": "team1", "foo": "bar"}',
            }
        }
    )
    executor = AWSBatchExecutor("batch", scheduler, config["batch"])
    executor._aws_user = "alice"

    @task(batch_tags={"step": "final", "project": "acme"}, namespace="test")
    def task1(x):
        return x

    exec1 = Execution("123")
    job = Job(task1, task1(10), execution=exec1)

    batch_tags = executor._get_job_options(job)["batch_tags"]
    assert batch_tags == {
        "redun_aws_user": "alice",
        "redun_execution_id": "123",
        "redun_job_id": job.id,
        "redun_project": "test",
        "redun_task_name": "test.task1",
        "step": "final",
        "team": "team1",
        "project": "acme",
        "foo": "bar",
    }


def test_batch_tags_no_default(scheduler: Scheduler) -> None:
    """
    Executor config should be able to turn off default batch tags.
    """
    # Setup executor.
    config = Config(
        {
            "batch": {
                "image": "image",
                "queue": "queue",
                "s3_scratch": "s3_scratch_prefix",
                "default_batch_tags": "false",
            }
        }
    )
    executor = AWSBatchExecutor("batch", scheduler, config["batch"])

    @task(batch_tags={"step": "final", "project": "acme"}, namespace="test")
    def task1(x):
        return x

    exec1 = Execution("123")
    job = Job(task1, task1(10), execution=exec1)

    batch_tags = executor._get_job_options(job)["batch_tags"]
    assert batch_tags == {
        "step": "final",
        "project": "acme",
    }


def test_executor_config(scheduler: Scheduler) -> None:
    """
    Executor should be able to parse its config.
    """
    # Setup executor.
    config = Config(
        {
            "batch": {
                "image": "image",
                "queue": "queue",
                "s3_scratch": "s3_scratch_prefix",
                "code_includes": "*.txt",
                "job_def_extra": '{"a_json_key": {"a_json_nested_key": 42}}',
            }
        }
    )
    executor = AWSBatchExecutor("batch", scheduler, config["batch"])

    assert executor.image == "image"
    assert executor.queue == "queue"
    assert executor.s3_scratch_prefix == "s3_scratch_prefix"
    assert isinstance(executor.code_package, dict)
    assert executor.code_package["includes"] == ["*.txt"]
    assert executor.debug is False
    assert executor.default_task_options["job_def_extra"] == {
        "a_json_key": {"a_json_nested_key": 42}
    }


@task()
def task1(x):
    return x + 10


@task(load_module="custom.module")
def task1_custom_module(x):
    return x + 10


@use_tempdir
@mock_s3
@patch("redun.executors.aws_batch.batch_submit")
@pytest.mark.parametrize(
    "custom_module, expected_load_module, a_task",
    [
        (None, "redun.tests.test_aws_batch", task1),
        ("custom.module", "custom.module", task1_custom_module),
    ],
)
def test_submit_task(batch_submit_mock, custom_module, expected_load_module, a_task):
    job_id = "123"
    image = "my-image"
    queue = "queue"
    s3_scratch_prefix = "s3://example-bucket/redun/"

    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")

    redun.executors.aws_batch.batch_submit.return_value = {"jobId": "batch-job-id"}

    # Create example workflow script to be packaged.
    File("workflow.py").write(
        f"""
@task(load_module={custom_module})
def task1(x):
    return x + 10
    """
    )

    job = Job(a_task, a_task())
    job.id = job_id
    job.eval_hash = "eval_hash"
    code_file = package_code(s3_scratch_prefix)
    resp = submit_task(
        image,
        queue,
        s3_scratch_prefix,
        job,
        a_task,
        args=[10],
        kwargs={},
        code_file=code_file,
    )

    # We should get a AWS Batch job id back.
    assert resp["jobId"] == "batch-job-id"

    # Input files should be made.
    assert File("s3://example-bucket/redun/jobs/eval_hash/input").exists()
    [code_file] = list(Dir("s3://example-bucket/redun/code"))

    # We should have submitted a job to AWS Batch.
    redun.executors.aws_batch.batch_submit.assert_called_with(
        {
            "containerOverrides": {
                "command": [
                    "redun",
                    "--check-version",
                    REDUN_REQUIRED_VERSION,
                    "oneshot",
                    expected_load_module,
                    "--code",
                    code_file.path,
                    "--input",
                    "s3://example-bucket/redun/jobs/eval_hash/input",
                    "--output",
                    "s3://example-bucket/redun/jobs/eval_hash/output",
                    "--error",
                    "s3://example-bucket/redun/jobs/eval_hash/error",
                    a_task.name,
                ]
            }
        },
        "queue",
        image="my-image",
        job_def_suffix="-redun-jd",
        job_name="batch-job-eval_hash",
        array_size=0,
        aws_region="us-west-2",
    )


@use_tempdir
@mock_s3
@patch("redun.executors.aws_batch.batch_submit")
@pytest.mark.parametrize(
    "custom_module, expected_load_module, a_task",
    [
        (None, "redun.tests.test_aws_batch", task1),
        ("custom.module", "custom.module", task1_custom_module),
    ],
)
def test_submit_task_array(batch_submit_mock, custom_module, expected_load_module, a_task):
    """
    oneshot should use --array-job when submitting an array job.
    """
    job_id = "123"
    image = "my-image"
    queue = "queue"
    s3_scratch_prefix = "s3://example-bucket/redun/"

    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")

    redun.executors.aws_batch.batch_submit.return_value = {"jobId": "batch-job-id"}

    # Create example workflow script to be packaged.
    File("workflow.py").write(
        f"""
@task(load_module={custom_module})
def task1(x):
    return x + 10
    """
    )

    job = Job(a_task, a_task())
    job.id = job_id
    job.eval_hash = "eval_hash"
    code_file = File("s3://example-bucket/redun/code/123.tar.gz")
    resp = submit_task(
        image,
        queue,
        s3_scratch_prefix,
        job,
        a_task,
        args=[10],
        kwargs={},
        code_file=code_file,
        array_uuid="123456",
        array_size=10,
    )

    # We should get a AWS Batch job id back.
    assert resp["jobId"] == "batch-job-id"

    # We should have submitted a job to AWS Batch.
    redun.executors.aws_batch.batch_submit.assert_called_with(
        {
            "containerOverrides": {
                "command": [
                    "redun",
                    "--check-version",
                    REDUN_REQUIRED_VERSION,
                    "oneshot",
                    expected_load_module,
                    "--code",
                    code_file.path,
                    "--array-job",
                    "--input",
                    "s3://example-bucket/redun/array_jobs/123456/input",
                    "--output",
                    "s3://example-bucket/redun/array_jobs/123456/output",
                    "--error",
                    "s3://example-bucket/redun/array_jobs/123456/error",
                    a_task.name,
                ]
            }
        },
        "queue",
        image="my-image",
        job_def_suffix="-redun-jd",
        job_name="batch-job-123456-array",
        array_size=10,
        aws_region="us-west-2",
    )


@use_tempdir
@mock_s3
@patch("redun.executors.aws_batch.batch_submit")
def test_submit_multinode_task(batch_submit_mock):
    job_id = "123"
    image = "my-image"
    queue = "queue"
    s3_scratch_prefix = "s3://example-bucket/redun/"

    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")

    redun.executors.aws_batch.batch_submit.return_value = {"jobId": "batch-job-id"}

    # Create example workflow script to be packaged.
    File("workflow.py").write(
        """
@task()
def task1(x):
    return x + 10
    """
    )

    job = Job(task1, task1())
    job.id = job_id
    job.eval_hash = "eval_hash"
    code_file = package_code(s3_scratch_prefix)
    resp = submit_task(
        image,
        queue,
        s3_scratch_prefix,
        job,
        task1,
        args=[10],
        kwargs={},
        code_file=code_file,
        job_options={"num_nodes": 4},
    )

    # We should get a AWS Batch job id back.
    assert resp["jobId"] == "batch-job-id"

    # Input files should be made.
    assert File("s3://example-bucket/redun/jobs/eval_hash/input").exists()
    [code_file] = list(Dir("s3://example-bucket/redun/code"))

    # We should have submitted a job to AWS Batch.
    redun.executors.aws_batch.batch_submit.assert_called_with(
        {
            "nodeOverrides": {
                "nodePropertyOverrides": [
                    {
                        "targetNodes": "0",
                        "containerOverrides": {
                            "command": [
                                "redun",
                                "--check-version",
                                REDUN_REQUIRED_VERSION,
                                "oneshot",
                                "redun.tests.test_aws_batch",
                                "--code",
                                code_file.path,
                                "--input",
                                "s3://example-bucket/redun/jobs/eval_hash/input",
                                "--output",
                                "s3://example-bucket/redun/jobs/eval_hash/output",
                                "--error",
                                "s3://example-bucket/redun/jobs/eval_hash/error",
                                task1.name,
                            ]
                        },
                    },
                    {
                        "targetNodes": "1:",
                        "containerOverrides": {
                            "command": [
                                "redun",
                                "--check-version",
                                REDUN_REQUIRED_VERSION,
                                "oneshot",
                                "redun.tests.test_aws_batch",
                                "--code",
                                code_file.path,
                                "--no-cache",
                                "--input",
                                "s3://example-bucket/redun/jobs/eval_hash/input",
                                "--output",
                                "/dev/null",
                                "--error",
                                "s3://example-bucket/redun/jobs/eval_hash/error",
                                task1.name,
                            ]
                        },
                    },
                ],
            }
        },
        "queue",
        image="my-image",
        job_def_suffix="-redun-jd",
        job_name="batch-job-eval_hash",
        array_size=0,
        aws_region="us-west-2",
        num_nodes=4,
    )


@use_tempdir
@mock_s3
@patch("redun.executors.aws_batch.batch_submit")
def test_submit_task_deep_file(batch_submit_mock):
    """
    Executor should be able to submit a task defined in a deeply nested file path.
    """
    job_id = "123"
    image = "my-image"
    queue = "queue"
    s3_scratch_prefix = "s3://example-bucket/redun/"

    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")

    redun.executors.aws_batch.batch_submit.return_value = {"jobId": "batch-job-id"}

    # Create example workflow script to be packaged.
    File("path/to/workflow.py").write(
        """
from redun import task

@task()
def task1(x):
    return x + 10
    """
    )

    module = import_script("path/to/workflow.py")

    job = Job(module.task1, module.task1())
    job.id = job_id
    job.eval_hash = "eval_hash"
    code_file = package_code(s3_scratch_prefix)
    resp = submit_task(
        image,
        queue,
        s3_scratch_prefix,
        job,
        module.task1,
        args=[10],
        kwargs={},
        code_file=code_file,
    )

    # We should get a AWS Batch job id back.
    assert resp["jobId"] == "batch-job-id"

    # Input files should be made.
    assert File("s3://example-bucket/redun/jobs/eval_hash/input").exists()
    [code_file] = list(Dir("s3://example-bucket/redun/code"))

    # We should have submitted a job to AWS Batch.
    redun.executors.aws_batch.batch_submit.assert_called_with(
        {
            "containerOverrides": {
                "command": [
                    "redun",
                    "--check-version",
                    REDUN_REQUIRED_VERSION,
                    "oneshot",
                    "workflow",
                    "--import-path",
                    "path/to",
                    "--code",
                    code_file.path,
                    "--input",
                    "s3://example-bucket/redun/jobs/eval_hash/input",
                    "--output",
                    "s3://example-bucket/redun/jobs/eval_hash/output",
                    "--error",
                    "s3://example-bucket/redun/jobs/eval_hash/error",
                    "task1",
                ]
            }
        },
        "queue",
        image="my-image",
        job_def_suffix="-redun-jd",
        job_name="batch-job-eval_hash",
        array_size=0,
        aws_region="us-west-2",
    )


@mock_s3
def test_parse_job_error() -> None:
    """
    We should be able to parse the error of a failed task.
    """
    s3_scratch_prefix = "s3://example-bucket/redun/"

    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")

    @task()
    def task1(x):
        return x + 1

    @task(script=True)
    def task_script1():
        return "echo hello!"

    expr = task1(10)
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"

    # Normal task, no error file.
    error, error_traceback = parse_job_error(s3_scratch_prefix, job)
    assert isinstance(error, ExceptionNotFoundError)
    assert isinstance(error_traceback, Traceback)

    # Simulate AWS Batch job failing.
    error = ValueError("Boom")
    error_file = File("s3://example-bucket/redun/jobs/eval_hash/error")
    error_file.write(pickle_dumps((error, Traceback.from_error(error))), mode="wb")

    # Normal task, error file exists.
    error, error_traceback = parse_job_error(s3_scratch_prefix, job)
    assert isinstance(error, ValueError)
    assert isinstance(error_traceback, Traceback)

    # Create a script task and job.
    expr2 = task_script1()
    job2 = Job(task_script1, expr2)
    job2.eval_hash = "eval_hash2"

    # Script task without an error file should retutn a generic error.
    error, error_traceback = parse_job_error(s3_scratch_prefix, job2)
    assert isinstance(error, ExceptionNotFoundError)
    assert isinstance(error_traceback, Traceback)

    # Create error file for script task.
    error_file2 = File("s3://example-bucket/redun/jobs/eval_hash2/error")
    error_file2.write("Boom")

    # Script task with an error file should return a specific error.
    error, error_traceback = parse_job_error(s3_scratch_prefix, job2)
    assert isinstance(error, ScriptError)
    assert error.message == "Boom"
    assert isinstance(error_traceback, Traceback)


@freeze_time("2020-10-13 06:47:14", tz_offset=-7)
@mock_logs
@patch("redun.executors.aws_batch.aws_describe_jobs")
def test_iter_batch_job_logs(aws_describe_jobs_mock) -> None:
    """
    We should be able to iterate through the logs of a Batch Job.
    """
    stream_name = "redun_aws_batch_example-redun-jd/default/6c939514f4054fdfb5ee65acc8aa4b07"
    aws_describe_jobs_mock.side_effect = lambda *args, **kwargs: iter(
        [
            {
                "container": {
                    "logStreamName": stream_name,
                }
            }
        ]
    )

    timestamp = int(time.time() * 1000)

    # Setup logs mocks.
    logs_client = boto3.client("logs", region_name="us-west-2")
    logs_client.create_log_group(logGroupName=BATCH_LOG_GROUP)
    logs_client.create_log_stream(logGroupName=BATCH_LOG_GROUP, logStreamName=stream_name)
    resp = logs_client.put_log_events(
        logGroupName=BATCH_LOG_GROUP,
        logStreamName=stream_name,
        logEvents=[
            {"timestamp": timestamp, "message": "A message 1."},
            {"timestamp": timestamp + 1, "message": "A message 2."},
        ],
    )
    resp = logs_client.put_log_events(
        logGroupName=BATCH_LOG_GROUP,
        logStreamName=stream_name,
        logEvents=[
            {"timestamp": timestamp + 2, "message": "A message 3."},
            {"timestamp": timestamp + 3, "message": "A message 4."},
        ],
        sequenceToken=resp["nextSequenceToken"],
    )

    expected_events = [
        {"message": "A message 1.", "timestamp": timestamp},
        {"message": "A message 2.", "timestamp": timestamp + 1},
        {"message": "A message 3.", "timestamp": timestamp + 2},
        {"message": "A message 4.", "timestamp": timestamp + 3},
    ]

    # Fetch log events.
    job_id = "123"
    events = iter_batch_job_logs(job_id, limit=1)
    event_list = [
        {"message": event["message"], "timestamp": event["timestamp"]} for event in events
    ]
    assert event_list == expected_events

    # Fetch log events in reverse.
    events = iter_batch_job_logs(job_id, limit=1, reverse=True)
    event_list = [
        {"message": event["message"], "timestamp": event["timestamp"]} for event in events
    ]
    assert event_list == list(reversed(expected_events))

    # Fetch log events in reverse with larger page size.
    events = iter_batch_job_logs(job_id, limit=2, reverse=True)
    event_list = [
        {"message": event["message"], "timestamp": event["timestamp"]} for event in events
    ]
    assert event_list == list(reversed(expected_events))

    # Fetch log lines.
    lines = list(iter_batch_job_log_lines(job_id))
    assert lines == [
        "2020-10-12 23:47:14  A message 1.",
        "2020-10-12 23:47:14.001000  A message 2.",
        "2020-10-12 23:47:14.002000  A message 3.",
        "2020-10-12 23:47:14.003000  A message 4.",
    ]

    # Fetch logs from unknown job.
    aws_describe_jobs_mock.side_effect = lambda *args, **kwargs: iter([])
    assert list(iter_batch_job_logs("unknown_job_id")) == []

    # Fetch logs from job with missing logs.
    aws_describe_jobs_mock.side_effect = lambda *args, **kwargs: iter(
        [
            {
                "container": {
                    "logStreamName": "bad_logs",
                }
            }
        ]
    )
    assert list(iter_batch_job_logs(job_id, required=False)) == []

    with pytest.raises(Exception):
        list(iter_batch_job_logs(job_id, required=True))

    # Fetch logs from job with no logs at all.
    aws_describe_jobs_mock.side_effect = lambda *args, **kwargs: iter([{"container": {}}])
    assert list(iter_batch_job_logs(job_id, required=False)) == []


def mock_executor(
    scheduler, custom_config_args: Dict[str, Any] = {}, debug=False, code_package=False
):
    """
    Returns an AWSBatchExecutor with AWS API mocks.
    """
    image = "my-image"
    queue = "queue"
    s3_scratch_prefix = "s3://example-bucket/redun/"

    # Setup executor.
    config = Config(
        {
            "batch": {
                "image": image,
                "queue": queue,
                "s3_scratch": s3_scratch_prefix,
                "job_monitor_interval": 0.05,
                "job_stale_time": 0.01,
                "code_package": code_package,
                "debug": debug,
                **custom_config_args,
            }
        }
    )
    executor = AWSBatchExecutor("batch", scheduler, config["batch"])

    executor.get_jobs = Mock()  # type:ignore[method-assign]
    executor.get_jobs.return_value = []

    executor.get_array_child_jobs = Mock()  # type:ignore[method-assign]
    executor.get_array_child_jobs.return_value = []

    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    return executor


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.parse_job_logs")
@patch("redun.executors.aws_batch.iter_batch_job_status")
@patch("redun.executors.aws_batch.batch_submit")
def test_executor(
    batch_submit_mock,
    iter_batch_job_status_mock,
    parse_job_logs_mock,
    get_aws_user_mock,
) -> None:
    """
    Ensure that we can submit job to AWSBatchExecutor.
    """
    batch_job_id = "batch-job-id"
    batch_job2_id = "batch-job2-id"

    # Setup AWS Batch mocks.
    iter_batch_job_status_mock.return_value = iter([])
    parse_job_logs_mock.return_value = []

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler, {"timeout": 5, "shared_memory": 5, "privileged": True})
    executor.start()

    try:
        batch_submit_mock.return_value = {
            "jobId": batch_job_id,
        }

        # Submit redun job that will succeed.
        expr = task1(10)
        job = Job(task1, expr)
        job.eval_hash = "eval_hash"
        job.args = ((10,), {})
        executor.submit(job)

        # Let job get stale so job arrayer actually submits it.
        wait_until(lambda: executor.arrayer.num_pending == 0)

        # Ensure job options were passed correctly.
        assert batch_submit_mock.call_args
        assert batch_submit_mock.call_args[1] == {
            "image": "my-image",
            "job_name": "redun-job-eval_hash",
            "job_def_suffix": "-redun-jd",
            "array_size": 0,
            "vcpus": 1,
            "gpus": 0,
            "memory": 4,
            "shared_memory": 5,
            "timeout": 5,
            "privileged": True,
            "autocreate_job_def": True,
            "job_def_name": None,
            "role": None,
            "retries": 1,
            "aws_region": "us-west-2",
            "num_nodes": None,
            "job_def_extra": None,
            "batch_tags": {
                "redun_aws_user": "alice",
                "redun_execution_id": "",
                "redun_job_id": job.id,
                "redun_project": "",
                "redun_task_name": "task1",
            },
            "share_id": None,
            "scheduling_priority_override": None,
        }

        batch_submit_mock.return_value = {
            "jobId": batch_job2_id,
        }

        # Submit redun job that will fail.
        expr2 = task1.options(memory=8)("a")
        job2 = Job(task1, expr2)
        job2.eval_hash = "eval_hash2"
        job2.args = (("a",), {})
        executor.submit(job2)

        # Let job get stale so job arrayer actually submits it.
        wait_until(lambda: executor.arrayer.num_pending == 0)

        # Ensure job options were passed correctly.
        assert batch_submit_mock.call_args[1] == {
            "image": "my-image",
            "job_name": "redun-job-eval_hash2",
            "job_def_suffix": "-redun-jd",
            "array_size": 0,
            "vcpus": 1,
            "gpus": 0,
            "memory": 8,
            "shared_memory": 5,
            "timeout": 5,
            "privileged": True,
            "autocreate_job_def": True,
            "job_def_name": None,
            "role": None,
            "retries": 1,
            "aws_region": "us-west-2",
            "num_nodes": None,
            "job_def_extra": None,
            "batch_tags": {
                "redun_aws_user": "alice",
                "redun_execution_id": "",
                "redun_job_id": job2.id,
                "redun_project": "",
                "redun_task_name": "task1",
            },
            "share_id": None,
            "scheduling_priority_override": None,
        }

        # Simulate AWS Batch completing job.
        output_file = File("s3://example-bucket/redun/jobs/eval_hash/output")
        output_file.write(pickle_dumps(task1.func(10)), mode="wb")

        # Simulate AWS Batch failing.
        error = ValueError("Boom")
        error_file = File("s3://example-bucket/redun/jobs/eval_hash2/error")
        error_file.write(pickle_dumps((error, Traceback.from_error(error))), mode="wb")

        iter_batch_job_status_mock.return_value = iter(
            [
                {
                    "jobId": batch_job_id,
                    "status": SUCCEEDED,
                    "container": {"logStreamName": "log1"},
                },
                {
                    "jobId": batch_job2_id,
                    "status": FAILED,
                    "container": {"logStreamName": "log2"},
                },
            ]
        )

        scheduler.batch_wait([job.id, job2.id])
        executor.stop()

        # Job results and errors should be sent back to scheduler.
        assert scheduler.job_results[job.id] == 20
        assert isinstance(scheduler.job_errors[job2.id], ValueError)

        # Assert job tags.
        job.job_tags == [("aws_batch_job", "batch-job-id"), ("aws_log_stream", "log1")]
        job.job_tags == [("aws_batch_job", "batch-job2-id"), ("aws_log_stream", "log2")]
    finally:
        # Ensure executor threads stop before continuing.
        executor.stop()


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.parse_job_logs")
@patch("redun.executors.aws_batch.iter_batch_job_status")
@patch("redun.executors.aws_batch.batch_submit")
def test_executor_multinode(
    batch_submit_mock,
    iter_batch_job_status_mock,
    parse_job_logs_mock,
    get_aws_user_mock,
) -> None:
    """
    Ensure that we can submit job to AWSBatchExecutor.
    """
    batch_job_id = "batch-job-id"

    # Setup AWS Batch mocks.
    iter_batch_job_status_mock.return_value = iter([])
    parse_job_logs_mock.return_value = []

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.start()

    try:
        batch_submit_mock.return_value = {
            "jobId": batch_job_id,
        }

        # Submit redun job that will succeed.
        expr = task1.options(num_nodes=4)(10)
        job = Job(task1, expr)
        job.eval_hash = "eval_hash"
        job.args = ((10,), {})
        executor.submit(job)

        # Let job get stale so job arrayer actually submits it.
        wait_until(lambda: executor.arrayer.num_pending == 0)

        # Ensure job options were passed correctly.
        assert batch_submit_mock.call_args
        assert batch_submit_mock.call_args[1] == {
            "image": "my-image",
            "job_name": "redun-job-eval_hash",
            "job_def_suffix": "-redun-jd",
            "array_size": 0,
            "vcpus": 1,
            "gpus": 0,
            "memory": 4,
            "shared_memory": None,
            "timeout": None,
            "privileged": False,
            "autocreate_job_def": True,
            "job_def_name": None,
            "role": None,
            "retries": 1,
            "aws_region": "us-west-2",
            "num_nodes": 4,
            "job_def_extra": None,
            "batch_tags": {
                "redun_aws_user": "alice",
                "redun_execution_id": "",
                "redun_job_id": job.id,
                "redun_project": "",
                "redun_task_name": "task1",
            },
            "share_id": None,
            "scheduling_priority_override": None,
        }
        job.job_tags == [
            ("aws_batch_job", "batch-job-id#0"),
            ("aws_log_stream", "log1"),
        ]
    finally:
        # Ensure executor threads stop before continuing.
        executor.stop()


@use_tempdir
@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.docker.iter_job_status")
@patch("redun.executors.docker.run_docker")
def test_executor_docker(
    run_docker_mock: Mock,
    iter_local_job_status_mock: Mock,
    get_aws_user_mock: Mock,
) -> None:
    """
    Ensure that we can submit job to AWSBatchExecutor with debug=True.
    """
    batch_job_id = "batch-job-id"
    batch_job2_id = "batch-job2-id"

    # Setup Docker mocks.
    iter_local_job_status_mock.return_value = iter([])

    # Setup redun mocks.
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler, debug=True)
    executor.start()

    try:
        run_docker_mock.return_value = batch_job_id

        # Submit redun job that will succeed.
        expr = task1(10)
        job = Job(task1, expr)
        job.eval_hash = "eval_hash"
        job.args = ((10,), {})
        executor.submit(job)

        # Let job get stale so job arrayer actually submits it.
        wait_until(lambda: executor.arrayer.num_pending == 0)

        # Ensure job options were passed correctly.
        scratch_dir = executor._docker_executor._scratch_prefix
        assert run_docker_mock.call_args[1] == {
            "image": "my-image",
            "gpus": 0,
            "memory": 4,
            "shared_memory": None,
            "vcpus": 1,
            "volumes": [(scratch_dir, scratch_dir)],
            "interactive": False,
            "include_aws_env": True,
        }

        run_docker_mock.reset_mock()
        run_docker_mock.return_value = batch_job2_id

        # Hand create Job and submit.
        expr2 = task1("a")
        job2 = Job(task1, expr2)
        job2.eval_hash = "eval_hash2"
        job2.args = (("a",), {})
        executor.submit(job2)

        # Let job get stale so job arrayer actually submits it.
        wait_until(lambda: executor.arrayer.num_pending == 0)

        # Ensure job options were passed correctly.
        assert run_docker_mock.call_args[1] == {
            "image": "my-image",
            "gpus": 0,
            "memory": 4,
            "shared_memory": None,
            "vcpus": 1,
            "volumes": [(scratch_dir, scratch_dir)],
            "interactive": False,
            "include_aws_env": True,
        }

        # Simulate output file created by job.
        output_file = File(f"{scratch_dir}/jobs/eval_hash/output")
        output_file.write(pickle_dumps(task1.func(10)), mode="wb")

        # Simulate AWS Batch failing.
        error = ValueError("Boom")
        error_file = File(f"{scratch_dir}/jobs/eval_hash2/error")
        error_file.write(pickle_dumps((error, Traceback.from_error(error))), mode="wb")

        iter_local_job_status_mock.return_value = iter(
            [
                {"jobId": batch_job_id, "status": SUCCEEDED, "logs": ""},
                {"jobId": batch_job2_id, "status": FAILED, "logs": ""},
            ]
        )

        scheduler.batch_wait([job.id, job2.id])

        # Job results and errors should be sent back to scheduler.
        assert scheduler.job_results[job.id] == 20
        assert isinstance(scheduler.job_errors[job2.id], ValueError)

    finally:
        # Ensure executor threads stop before continuing.
        executor.stop()


@use_tempdir
@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.docker.iter_job_status")
@patch("redun.executors.docker.run_docker")
def test_executor_job_debug(
    run_docker_mock: Mock,
    iter_local_job_status_mock: Mock,
    get_aws_user_mock: Mock,
) -> None:
    """
    Ensure that we can submit job to AWSBatchExecutor with debug=True in task options.
    """
    batch_job_id = "batch-job-id"

    # Setup Docker mocks.
    iter_local_job_status_mock.return_value = iter([])

    # Setup redun mocks.
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.start()

    try:
        run_docker_mock.return_value = batch_job_id

        # Submit redun job that will succeed.
        expr = task1.options(debug=True)(10)
        job = Job(task1, expr)
        job.eval_hash = "eval_hash"
        job.args = ((10,), {})
        executor.submit(job)

        # Let job get stale so job arrayer actually submits it.
        wait_until(lambda: executor.arrayer.num_pending == 0)

        # Ensure job options were passed correctly.
        scratch_dir = executor._docker_executor._scratch_prefix
        assert run_docker_mock.call_args[1] == {
            "image": "my-image",
            "gpus": 0,
            "memory": 4,
            "shared_memory": None,
            "vcpus": 1,
            "volumes": [(scratch_dir, scratch_dir)],
            "interactive": False,
            "include_aws_env": True,
        }

        # Simulate output file created by job.
        output_file = File(f"{scratch_dir}/jobs/eval_hash/output")
        output_file.write(pickle_dumps(task1.func(10)), mode="wb")

        iter_local_job_status_mock.return_value = iter(
            [
                {"jobId": batch_job_id, "status": SUCCEEDED, "logs": ""},
            ]
        )

        scheduler.batch_wait([job.id])

        # Job results and errors should be sent back to scheduler.
        assert scheduler.job_results[job.id] == 20

    finally:
        # Ensure executor threads stop before continuing.
        executor.stop()


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.parse_job_logs")
@patch("redun.executors.aws_batch.iter_batch_job_status")
@patch("redun.executors.aws_batch.batch_submit")
def test_executor_error_override(
    batch_submit_mock: Mock,
    iter_batch_job_status_mock: Mock,
    parse_job_logs_mock: Mock,
    get_aws_user_mock: Mock,
) -> None:
    """
    Some AWS Batch errors should be overridden.
    """

    @task()
    def task1(x):
        return x + 10

    @task(script=True)
    def task_script1(x):
        return "ls"

    batch_job_id = "batch-job-id"
    batch_job_script_id = "batch-job-script-id"

    # Setup AWS Batch mocks.
    iter_batch_job_status_mock.return_value = iter([])
    parse_job_logs_mock.return_value = []

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.start()

    batch_submit_mock.return_value = {
        "jobId": batch_job_id,
    }

    # Submit redun job that will succeed at the redun-level.
    expr = task1.options(memory=8)("a")
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"
    job.args = (("a",), {})
    executor.submit(job)

    # Let job get stale so job arrayer actually submits it.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    batch_submit_mock.return_value = {
        "jobId": batch_job_script_id,
    }

    # Submit redun job that will succeed at the redun-level.
    expr_script = task_script1.options(memory=8)("a")
    job_script = Job(task_script1, expr_script)
    job_script.eval_hash = "eval_script_hash"
    job_script.args = (("a",), {})
    executor.submit(job_script)

    # Let job get stale so job arrayer actually submits it.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Simulate output file created by job.
    output_file = File("s3://example-bucket/redun/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Simulate output file created by job.
    output_file = File("s3://example-bucket/redun/jobs/eval_script_hash/output")
    output_file.write(pickle_dumps("done"), mode="wb")
    File("s3://example-bucket/redun/jobs/eval_script_hash/status").write("ok")

    # But simulate AWS Batch failing.
    reason = "CannotInspectContainerError: Could not transition to inspecting."
    iter_batch_job_status_mock.return_value = iter(
        [
            {
                "jobId": batch_job_id,
                "status": FAILED,
                "attempts": [
                    {
                        "container": {
                            "reason": reason,
                        },
                    },
                ],
            },
            {
                "jobId": batch_job_script_id,
                "status": FAILED,
                "attempts": [
                    {
                        "container": {
                            "reason": reason,
                        },
                    },
                ],
            },
        ]
    )

    scheduler.batch_wait([job.id, job_script.id])
    executor.stop()

    # Despite AWS Batch error, redun job should succeed and
    # results should be sent back to scheduler.
    assert scheduler.job_results[job.id] == 20


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.docker.iter_job_status")
@patch("redun.executors.docker.run_docker")
def test_executor_multiple_start(
    run_docker_mock, iter_local_job_status_mock, get_aws_user_mock
) -> None:
    """
    Ensure that we can start executor multiple times.
    """
    # Setup Docker mocks.
    iter_local_job_status_mock.return_value = iter([])

    # Setup redun mocks.
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler, debug=True)
    executor.start()
    executor._start()
    executor._start()
    executor.stop()


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.docker.iter_job_status")
@patch("redun.executors.docker.run_docker")
def test_interactive(run_docker_mock, iter_local_job_status_mock, get_aws_user_mock) -> None:
    """
    The interactive task option should be passed to run_docker.
    """
    # Setup Docker mocks.
    iter_local_job_status_mock.return_value = iter([])
    run_docker_mock.return_value = "batch-job-id"

    # Setup redun mocks.
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler, debug=True)
    executor.start()

    # Hand create Job and submit.
    expr = task1.options(interactive=True)(10)
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"
    job.args = ((10,), {})
    executor.submit(job)

    # Let job get stale so job arrayer actually submits it
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Ensure job options were passed correctly.
    scratch_dir = executor._docker_executor._scratch_prefix
    assert run_docker_mock.call_args[1] == {
        "image": "my-image",
        "interactive": True,
        "gpus": 0,
        "memory": 4,
        "shared_memory": None,
        "vcpus": 1,
        "volumes": [(scratch_dir, scratch_dir)],
        "include_aws_env": True,
    }

    # Cleanly stop executor.
    executor.stop()


@mock_s3
def test_executor_handles_unrelated_jobs() -> None:
    """
    Regression test for https://insitro.atlassian.net/browse/DE-2632

    There is an expanding pattern of using a "headnode" running in batch to trigger redun
    pipelines. If the headnode and redun jobs that it spawns have a shared job_name_prefix then
    the headnode job can get gathered in the `get_jobs` call and we will try to extract the hash.
    However, since the headnode job is not a redun job, it will not have a hash and previously
    caused execution failures.

    This test confirms that jobs without hashes in their names are ignored which allows headnode
    jobs(triggered via lambda or otherwise) to share job name prefixes with the redun jobs that
    they spawn.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)

    prefix = "liveratlas_spearmancor"
    hash1 = "123456789"
    hash2 = "987654321"

    # Set up mocks to include a headnode job(no hash) and some redun jobs that it "spawned".
    executor.get_jobs.return_value = [
        # The headnode job. Note the job name has not hash in it as the hash appears after the "-"
        # in a redun job name.
        {"jobId": "headnode", "jobName": f"{prefix}_automation_headnode"},
        # Redun jobs that were triggered by the "redun run" in the headnode.
        {"jobId": "preprocess", "jobName": f"{prefix}_preprocess-{hash1}"},
        {"jobId": "decode", "jobName": f"{prefix}_decode-{hash2}"},
    ]

    executor.gather_inflight_jobs()

    assert executor.preexisting_batch_jobs == {
        hash1: "preprocess",
        hash2: "decode",
    }


@mock_s3
def test_executor_inflight_array_job() -> None:
    """
    Ensure we reunite with an inflight array job
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)

    # Set up mocks to indicate an array job is inflight.
    array_uuid = str(uuid.uuid4()).replace("-", "")
    executor.get_jobs.return_value = [
        {"jobId": "carrots", "jobName": f"redun-job-{array_uuid}-array"}
    ]
    executor.get_array_child_jobs.return_value = [
        {"jobId": "carrots:1", "arrayProperties": {"index": 1}},
        {"jobId": "carrots:0", "arrayProperties": {"index": 0}},
        {"jobId": "carrots:2", "arrayProperties": {"index": 2}},
    ]

    # Set up hash scratch file
    eval_file = File(f"s3://example-bucket/redun/array_jobs/{array_uuid}/eval_hashes")
    with eval_file.open("w") as eval_f:
        eval_f.write("zero\none\ntwo")

    # Force the scheduler to gather the inflight jobs. This normally happens on
    # first job submission but we want to just check that we can join here.
    executor.gather_inflight_jobs()

    # Check query for child job happened.
    assert executor.get_array_child_jobs.call_args
    assert executor.get_array_child_jobs.call_args[0] == (
        "carrots",
        BATCH_JOB_STATUSES.inflight,
    )

    # Make sure child jobs (and not parent) ended up in pending batch jobs.
    assert executor.preexisting_batch_jobs == {
        "zero": "carrots:0",
        "one": "carrots:1",
        "two": "carrots:2",
    }


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.package_code")
@patch("redun.executors.aws_batch.iter_batch_job_status")
@patch("redun.executors.aws_batch.submit_task")
def test_code_packaging(
    submit_task_mock, iter_batch_job_status_mock, package_code_mock, get_aws_user_mock
) -> None:
    """
    Ensure that code packaging only happens on first submission.
    """
    package_code_mock.return_value = File("s3://fake-bucket/code.tar.gz")

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler, code_package=True)
    executor.start()

    # Starting the executor should not have triggered code packaging.
    assert executor.code_file is None
    assert package_code_mock.call_count == 0

    # Hand create jobs.
    job1 = Job(task1, task1(10))
    job1.id = "1"
    job1.task = task1
    job1.eval_hash = "eval_hash"
    job1.args = ((10,), {})

    job2 = Job(task1, task1(20))
    job2.id = "2"
    job2.eval_hash = "eval_hash"
    job2.args = ((20,), {})

    # Submit a job and ensure that the code was packaged.
    executor.submit(job1)
    assert executor.code_file
    assert executor.code_file.path == "s3://fake-bucket/code.tar.gz"
    assert package_code_mock.call_count == 1

    # Submit another job and ensure that code was not packaged again.
    executor.submit(job2)
    assert package_code_mock.call_count == 1

    executor.stop()


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.aws_describe_jobs")
def test_inflight_join_only_on_first_submission(aws_describe_jobs_mock, get_aws_user_mock) -> None:
    """
    Ensure that inflight jobs are only gathered once and not on every job submission.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)

    executor.start()

    # Hand create jobs.
    job1 = Job(task1, task1(10))
    job1.id = "1"
    job1.eval_hash = "eval_hash"
    job1.args = ((10,), {})

    job2 = Job(task1, task1(20))
    job2.id = "2"
    job2.eval_hash = "eval_hash"
    job2.args = ((20,), {})

    # Submit redun job.
    executor.submit(job1)

    # Ensure that inflight jobs were gathered.
    assert executor.get_jobs.call_count == 1

    # Submit the second job and confirm that job reuniting was not done again.
    executor.submit(job2)
    assert executor.get_jobs.call_count == 1

    executor.stop()


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.aws_describe_jobs")
@patch("redun.executors.aws_batch.iter_batch_job_status")
@patch("redun.executors.aws_batch.batch_submit")
def test_executor_inflight_job(
    batch_submit_mock,
    iter_batch_job_status_mock,
    aws_describe_jobs_mock,
    get_aws_user_mock,
) -> None:
    """
    Ensure we reunite with an inflight job.
    """
    batch_job_id = "333"

    # Setup AWS Batch mocks.
    iter_batch_job_status_mock.return_value = iter([])
    aws_describe_jobs_mock.return_value = iter(
        [
            {
                "jobId": batch_job_id,
            }
        ]
    )

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.get_jobs.return_value = [{"jobId": batch_job_id, "jobName": "redun-job-eval_hash"}]
    executor.start()

    # Hand create job.
    job = Job(task1, task1(10))
    job.id = "123"
    job.eval_hash = "eval_hash"
    job.args = ((10,), {})

    # Submit redun job.
    executor.submit(job)

    # Ensure no batch jobs were submitted.
    assert batch_submit_mock.call_count == 0

    # Simulate AWS Batch completing with valid value.
    output_file = File("s3://example-bucket/redun/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    iter_batch_job_status_mock.return_value = iter([{"jobId": batch_job_id, "status": SUCCEEDED}])

    scheduler.batch_wait([job.id])

    # Simulate pre-existing job output.
    output_file = File("s3://example-bucket/redun/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Ensure redun job is completed.
    assert scheduler.job_results[job.id] == 20

    executor.stop()


@task(limits={"cpu": 1}, random_option=5)
def array_task(x):
    return x + 10


@task()
def other_task(x, y):
    return x - y


# Tests begin here
def test_job_descrs() -> None:
    """Tests the JobDescription class used to determine if Jobs are equivalent"""
    j1 = Job(array_task, array_task(1))
    j2 = Job(array_task, array_task(2))

    a = job_array.JobDescription(j1)
    b = job_array.JobDescription(j2)

    assert hash(a) == hash(b)
    assert a == b

    j3 = Job(other_task, other_task(1, y=2))
    c = job_array.JobDescription(j3)
    assert a != c


@mock_s3
def test_job_staleness() -> None:
    """Tests staleness criteria for array'ing jobs"""
    j1 = Job(array_task, array_task(1))
    j1.args = ((1,), {})
    d = job_array.JobDescription(j1)

    sched = mock_scheduler()
    exec = mock_executor(sched)
    arr = job_array.JobArrayer(
        exec._submit_jobs,
        exec._on_error,
        submit_interval=10000.0,
        stale_time=0.05,
        min_array_size=5,
    )

    for i in range(10):
        arr.add_job(j1)

    assert arr.get_stale_descrs() == []
    wait_until(lambda: arr.get_stale_descrs() == [d])


@mock_s3
def test_arrayer_thread() -> None:
    """Tests that the arrayer monitor thread can be restarted after exit"""
    j1 = Job(array_task, array_task(1))
    j1.args = ((1,), {})

    sched = mock_scheduler()
    exec = mock_executor(sched)
    arr = job_array.JobArrayer(
        exec._submit_jobs,
        exec._on_error,
        submit_interval=10000.0,
        stale_time=0.05,
        min_array_size=5,
    )

    arr.add_job(j1)
    assert arr._monitor_thread.is_alive()

    # Stop the monitoring thread.
    arr.stop()
    assert not arr._monitor_thread.is_alive()

    # Submitting an additional job should restart the thread.
    arr.add_job(j1)
    assert arr._monitor_thread.is_alive()

    arr.stop()


@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.aws_describe_jobs")
@patch("redun.executors.aws_batch.submit_task")
def test_jobs_are_arrayed(
    submit_task_mock: Mock, aws_describe_jobs_mock: Mock, get_aws_user_mock: Mock
) -> None:
    """
    Tests repeated jobs are submitted as a single array job. Checks that
    job ID for the array job and child jobs end up tracked
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.arrayer.min_array_size = 3
    executor.arrayer.max_array_size = 7

    aws_describe_jobs_mock.return_value = iter([])
    submit_task_mock.side_effect = [
        {"jobId": "first-array-job", "arrayProperties": {"size": 7}},
        {"jobId": "second-array-job", "arrayProperties": {"size": 3}},
        {"jobId": "single-job"},
    ]

    test_jobs = []
    for i in range(10):
        job = Job(array_task, array_task(i))
        job.id = f"task_{i}"
        job.eval_hash = f"eval_hash_{i}"
        job.args = ((i,), {})

        executor.submit(job)
        test_jobs.append(job)

    # Wait for jobs to get submitted from arrayer to executor.
    wait_until(lambda: len(executor.pending_batch_jobs) == 10)

    # Two array jobs, of size 7 and 3, should have been submitted.
    pending_correct = {
        f"first-array-job:{i}": test_jobs[i] for i in range(executor.arrayer.max_array_size)
    }
    pending_correct.update(
        {
            f"second-array-job:{i}": j
            for i, j in enumerate(test_jobs[executor.arrayer.max_array_size :])
        }
    )
    assert executor.pending_batch_jobs == pending_correct

    # Two array jobs should have been submitted
    assert submit_task_mock.call_count == 2

    # Submit a different kind of job now.
    j = Job(other_task, other_task(3, 5))
    j.id = "other_task"
    j.eval_hash = "hashbrowns"
    j.args = ((3, 5), {})
    executor.submit(j)

    assert len(executor.arrayer.pending) == 1
    pending_correct["single-job"] = j
    wait_until(lambda: executor.pending_batch_jobs == pending_correct)

    # Make monitor thread exit correctly
    executor.stop()


@use_tempdir
@mock_s3
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.aws_batch.AWSBatchExecutor._submit_single_job")
def test_array_disabling(submit_single_mock, get_aws_user_mock):
    """
    Tests setting `min_array_size=0` disables job arraying.
    """
    # Setup executor.
    config = Config(
        {
            "batch": {
                "image": "image",
                "queue": "queue",
                "s3_scratch": "s3_scratch_prefix",
                "code_includes": "*.txt",
                "min_array_size": 0,
            }
        }
    )
    scheduler = mock_scheduler()

    executor = AWSBatchExecutor("batch", scheduler, config["batch"])
    executor.get_jobs = Mock()
    executor.get_jobs.return_value = []

    # Submit one test job.
    job = Job(other_task, other_task(5, 3))
    job.id = "carrots"
    job.eval_hash = "why do i always say carrots in test cases idk"
    job.args = ((5, 3), {})
    executor.submit(job)

    # Job should be submitted immediately.
    assert submit_single_mock.call_args
    assert submit_single_mock.call_args[0] == (job,)

    # Monitor thread should not run.
    assert not executor.arrayer._monitor_thread.is_alive()
    executor.stop()


@mock_s3
@use_tempdir
@patch("redun.executors.aws_batch.batch_submit")
def test_array_job_s3_setup(batch_submit_mock: Mock) -> None:
    """
    Tests that args, kwargs, and output file paths end up
    in the correct locations in S3 as the right data structure
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.s3_scratch_prefix = "./evil\ndirectory"

    batch_submit_mock.return_value = {
        "jobId": "array-job-id",
        "arrayProperties": {"size": "10"},
    }

    test_jobs = []
    for i in range(10):
        job = Job(other_task, other_task(i, y=2 * i))
        job.id = f"task_{i}"
        job.eval_hash = f"hash_{i}"
        job.args = ((i,), {"y": 2 * i})
        test_jobs.append(job)

    array_uuid = executor._submit_array_job(test_jobs)

    # Check input file is on S3 and contains list of (args, kwargs) tuples
    input_file = File(
        get_array_scratch_file(executor.s3_scratch_prefix, array_uuid, SCRATCH_INPUT)
    )
    assert input_file.exists()

    with input_file.open("rb") as infile:
        arglist, kwarglist = pickle.load(infile)
    assert arglist == [(i,) for i in range(10)]
    assert kwarglist == [{"y": 2 * i} for i in range(10)]

    # Check output paths file is on S3 and contains correct output paths
    output_file = File(
        get_array_scratch_file(executor.s3_scratch_prefix, array_uuid, SCRATCH_OUTPUT)
    )
    assert output_file.exists()
    ofiles = json.loads(output_file.read())

    assert ofiles == [
        get_job_scratch_file(executor.s3_scratch_prefix, j, SCRATCH_OUTPUT) for j in test_jobs
    ]

    # Error paths are the same as output, basically
    error_file = File(
        get_array_scratch_file(executor.s3_scratch_prefix, array_uuid, SCRATCH_ERROR)
    )
    assert error_file.exists()
    efiles = json.loads(error_file.read())

    assert efiles == [
        get_job_scratch_file(executor.s3_scratch_prefix, j, SCRATCH_ERROR) for j in test_jobs
    ]

    # Child job eval hashes should be present as well.
    eval_file = File(
        get_array_scratch_file(executor.s3_scratch_prefix, array_uuid, SCRATCH_HASHES)
    )
    with eval_file.open("r") as evfile:
        hashes = evfile.read().splitlines()

    assert hashes == [job.eval_hash for job in test_jobs]

    # Make monitor thread exit correctly
    executor.stop()


@mock_s3
@use_tempdir
@patch("redun.executors.aws_batch.batch_submit")
def test_array_oneshot(batch_submit_mock: Mock) -> None:
    """
    Checks array child jobs can fetch their args and kwargs, and
    put their (correct) output in the right place.
    """
    # Create a code file
    file = File("workflow.py")
    file.write(
        """
from redun import task

@task()
def other_task(x, y):
   return x / y
        """
    )
    create_tar("code.tar.gz", ["workflow.py"])
    file.remove()

    # Submit 10 jobs that will be arrayed
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.s3_scratch_prefix = "."

    batch_submit_mock.return_value = {
        "jobId": "array-job-id",
        "arrayProperties": {"size": "10"},
    }

    test_jobs = []
    for i in range(3):
        job = Job(other_task, other_task(i, y=2 * i))
        job.id = f"task_{i}"
        job.eval_hash = f"hash_{i}"
        job.args = ((i,), {"y": 2 * i})
        test_jobs.append(job)

    array_uuid = executor._submit_array_job(test_jobs)

    # Now run 2 of those jobs and make sure they work ok
    client = RedunClient()
    array_dir = os.path.join(executor.s3_scratch_prefix, "array_jobs", array_uuid)
    input_path = os.path.join(array_dir, SCRATCH_INPUT)
    output_path = os.path.join(array_dir, SCRATCH_OUTPUT)
    error_path = os.path.join(array_dir, SCRATCH_ERROR)
    executor.stop()

    for i in range(3):
        os.environ[job_array.AWS_ARRAY_VAR] = str(i)
        try:
            client.execute(
                [
                    "redun",
                    "oneshot",
                    "workflow.py",
                    "--code",
                    "code.tar.gz",
                    "--array-job",
                    "--input",
                    input_path,
                    "--output",
                    output_path,
                    "--error",
                    error_path,
                    "other_task",
                ]
            )

            # Check expected output.
            output_file = File(
                get_job_scratch_file(executor.s3_scratch_prefix, test_jobs[i], SCRATCH_OUTPUT)
            )
            assert pickle.loads(cast(bytes, output_file.read("rb"))) == i / (2 * i)
        except Exception:
            # Check expected error.
            error_file = File(
                get_job_scratch_file(executor.s3_scratch_prefix, test_jobs[i], SCRATCH_ERROR)
            )
            error, _ = pickle.loads(cast(bytes, error_file.read("rb")))
            assert isinstance(error, ZeroDivisionError)


@patch("redun.executors.aws_batch.aws_describe_jobs")
def test_get_log_stream(aws_describe_jobs_mock) -> None:
    """Test the log stream fetching utility, which handles the difference in behavior for
    multi-node."""

    aws_region = "fake_region"

    # Heavily pruned `JobDetail` messages
    single_node_status = {
        "jobArn": "arn:aws:batch:us-west-2:298579124006:job/89ee416c",
        "jobName": "redun-testing-c813121061b56b5934d5db78730ada2e2ae11e44",
        "jobId": "single_node_id",
        "status": "FAILED",
        "container": {
            "logStreamName": "log_stream_arn",
        },
        "nodeDetails": {"nodeIndex": 0, "isMainNode": True},
    }

    # This is the overall message for a multi-node job. There is no log stream, but there is
    # a `nodeProperties` field.
    multi_node_status = {
        "jobArn": "arn:aws:batch:us-west-2:298579124006:job/89ee416c-1ab1-457c-b728-1f10318f61bb",
        "jobName": "redun-testing-c813121061b56b5934d5db78730ada2e2ae11e44",
        "jobId": "multi_node_id",
        "attempts": [
            {
                "container": {
                    "exitCode": 1,
                    "logStreamName": "cannot_be_returned",
                    "networkInterfaces": [],
                },
                "startedAt": 1663274825002,
                "stoppedAt": 1663274840060,
                "statusReason": "Essential container in task exited",
            }
        ],
        "createdAt": 1663274377312,
        "retryStrategy": {"attempts": 2, "evaluateOnExit": []},
        "startedAt": 1663274957408,
        "parameters": {},
        "nodeProperties": {
            "numNodes": 3,
            "mainNode": 0,
            "nodeRangeProperties": [
                {
                    "targetNodes": "0",
                    "container": {
                        "image": "image_arn",
                    },
                },
                {
                    "targetNodes": "1:",
                    "container": {
                        "image": "image_arn",
                    },
                },
            ],
        },
    }

    # No query is required for single node
    aws_describe_jobs_mock.return_value = None
    assert get_job_log_stream(single_node_status, aws_region) == "log_stream_arn"

    # Multi node will query again and fetch from there.
    aws_describe_jobs_mock.return_value = iter([single_node_status])
    assert get_job_log_stream(multi_node_status, aws_region) == "log_stream_arn"
    assert aws_describe_jobs_mock.call_args == unittest.mock.call(
        ["multi_node_id#0"], aws_region=aws_region
    )


def test_equiv_job_def() -> None:
    """Check our equality testing for job definitions."""

    basic_job = get_job_details(image="some_image", role="some_role")
    # This is derived from a real query for a job description from batch, lightly redacted.
    # It has lots of keys we don't set, but shouldn't cause errors.
    existing_job = {
        "jobDefinitionName": "jobDefinitionName",
        "jobDefinitionArn": "jobDefinitionArn",
        "revision": 12345,
        "status": "ACTIVE",
        "type": "container",
        "parameters": {},
        "containerProperties": {
            "image": "some_image",
            "vcpus": 1,
            "memory": 4,
            "command": ["ls"],
            "jobRoleArn": "some_role",
            "volumes": [],
            "environment": [],
            "mountPoints": [],
            "privileged": False,
            "ulimits": [],
            "resourceRequirements": [],
            "secrets": [],
        },
        "tags": {},
    }

    assert equiv_job_def(basic_job, existing_job)


################################################################################################
# Fair Share Scheduling related tests
################################################################################################


def test_fss_executor_config(scheduler: Scheduler) -> None:
    """
    FSS config variables flow are recognized by Executor.
    """
    # Setup executor.
    config = Config(
        {
            "batch": {
                "image": "image",
                "queue": "queue",
                "s3_scratch": "s3_scratch_prefix",
                "share_id": "team1",
                "scheduling_priority_override": 20,
            }
        }
    )
    executor = AWSBatchExecutor("batch", scheduler, config["batch"])

    assert executor.image == "image"
    assert executor.queue == "queue"
    assert executor.s3_scratch_prefix == "s3_scratch_prefix"
    assert isinstance(executor.code_package, dict)
    assert executor.debug is False

    @task(namespace="test")
    def task1(x):
        return x

    exec1 = Execution("123")
    job = Job(task1, task1(10), execution=exec1)

    options = executor._get_job_options(job)
    assert options["share_id"] == "team1"
    assert options["scheduling_priority_override"] == 20


def test_fss_task_override(scheduler: Scheduler) -> None:
    """
    FSS config variables can be overridden on Task
    """
    # Setup executor.
    config = Config(
        {
            "batch": {
                "image": "image",
                "queue": "queue",
                "s3_scratch": "s3_scratch_prefix",
                "share_id": "team1",
                "scheduling_priority_override": 20,
            }
        }
    )
    executor = AWSBatchExecutor("batch", scheduler, config["batch"])
    executor._aws_user = "alice"

    @task(share_id="subteam2", scheduling_priority_override=99, namespace="test")
    def task1(x):
        return x

    exec1 = Execution("123")
    job = Job(task1, task1(10), execution=exec1)

    options = executor._get_job_options(job)
    assert options["share_id"] == "subteam2"
    assert options["scheduling_priority_override"] == 99



import logging
import os
import sys
import time
from typing import Iterator, Optional, Tuple, cast

import pytest
from _pytest.monkeypatch import MonkeyPatch
from sqlalchemy import create_engine, event, text
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import Session

from redun import Scheduler
from redun.backends.db import RedunBackendDb, RedunSession
from redun.config import Config
from redun.task import get_task_registry
from redun.utils import clear_import_paths

logger = logging.getLogger(__name__)


def connect_with_retries(
    uri: str, interval: float = 0.1, timeout: float = 10.0
) -> Tuple[Engine, Connection]:
    """
    Connect to database with retries.
    """
    logger.info("Connecting to test database ({})...".format(uri))
    credentialed_uri = RedunBackendDb._get_credentialed_uri(uri, {})
    engine = create_engine(credentialed_uri, future=True)
    stop = time.time() + timeout
    while time.time() < stop:
        try:
            conn = engine.connect()
            logger.info("Connected to test database ({})".format(uri))
            return engine, conn
        except OperationalError:
            time.sleep(interval)

    raise RuntimeError("Unable to connect to test database: {}".format(uri))


@pytest.fixture(scope="session")
def monkeysession(request):
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


@pytest.fixture(scope="session")
def testdb(monkeysession) -> Iterator[Optional[str]]:
    """
    Create a test database.

    By using scope=session the test database is created only once per testing
    session.
    """
    host = "localhost"
    port = 5432
    database_name = "redun"

    monkeysession.setenv("REDUN_DB_USERNAME", "postgres")
    monkeysession.setenv("REDUN_DB_PASSWORD", "postgres")

    test_db_uri_stub = "postgresql://{host}:{port}/".format(
        host=host,
        port=port,
    )
    test_db_uri = test_db_uri_stub + database_name

    if not os.environ.get("REDUN_TEST_POSTGRES"):
        # Postgres database disabled.
        yield None
        return

    # Wait for database to accept connections.
    engine, conn = connect_with_retries(test_db_uri_stub)

    conn = conn.execution_options(isolation_level="AUTOCOMMIT")

    # Create test database.
    conn.execute(text("drop database if exists {}".format(database_name)))
    conn.execute(text("create database {}".format(database_name)))
    conn.close()
    engine.dispose()

    yield test_db_uri


@pytest.fixture
def scheduler(testdb: Optional[str]) -> Iterator[Scheduler]:
    """
    Returns a Scheduler for testing.
    """
    config: Optional[Config]

    if testdb:
        config = Config(
            {
                "scheduler": {"ignore_warnings": "namespace"},
                "backend": {"db_uri": testdb, "automigrate": "True"},
            }
        )
        scheduler = Scheduler(config=config)
        scheduler.load()
        backend = cast(RedunBackendDb, scheduler.backend)
        assert backend.engine and backend.session

        # We need to create a session bound to connection,
        # so let's get rid of the default one
        # background: https://docs.sqlalchemy.org/en/14/orm/session_transaction.html
        backend.session.close()

        # Run tests inside a transaction, so that we rollback effects.
        # start the session in a SAVEPOINT...
        connection = backend.engine.connect()

        trans = connection.begin()
        session = RedunSession(bind=connection, backend=backend)
        backend.session = session

        nested = connection.begin_nested()

        # then each time that SAVEPOINT ends, reopen it
        @event.listens_for(session, "after_transaction_end")
        def restart_savepoint(session, transaction):
            nonlocal nested
            if not nested.is_active:
                nested = connection.begin_nested()

        yield scheduler

        backend.session.close()
        trans.rollback()

        # Explicitly dispose the db engine so we can force kill the docker
        # process without error messages.
        connection.close()
        backend.engine.dispose()
        del scheduler

    else:
        yield Scheduler(
            config=Config(
                {
                    "scheduler": {"ignore_warnings": "namespace"},
                }
            )
        )


@pytest.fixture
def backend(scheduler: Scheduler) -> RedunBackendDb:
    """
    Returns a RedunBackend cast as RedunBackendDb.
    """
    backend = cast(RedunBackendDb, scheduler.backend)
    return backend


@pytest.fixture
def session(backend: RedunBackendDb) -> Session:
    """
    Returns a sqlalchemy Session for the redun backend.
    """
    assert backend.session
    return backend.session


@pytest.fixture(autouse=True)
def redun_globals(request):
    """
    pytest fixture for resetting redun global state during tests.

    - python modules
    - python sys.path
    - redun tasks
    - redun import paths
    """
    init_module_names = set(sys.modules.keys())
    init_sys_path = list(sys.path)
    task_registry = get_task_registry()

    # Temporarily remove unrelated tests from the task registry to prevent interference
    # (Note: Resetting the task registry to its original state at the end of this function doesn't
    # work because pytest loads all code modules before running any tests which means at this
    # point, all tasks across all unit tests are present in the task registry.)
    tasks_to_hide = {}
    for task_fullname in list(task_registry._tasks):
        if task_fullname.startswith("redun."):
            continue
        if task_fullname.startswith(request.module.__name__):
            continue
        if hasattr(request.module, "redun_namespace") and task_fullname.startswith(
            request.module.redun_namespace
        ):
            continue
        tasks_to_hide[task_fullname] = task_registry.get(task_fullname)
        del task_registry._tasks[task_fullname]

    yield

    # Reset imported modules.
    for module_name in list(sys.modules.keys()):
        if module_name not in init_module_names:
            sys.modules.pop(module_name)

    # Reset sys.path.
    sys.path = init_sys_path

    # Reset redun extra import paths.
    clear_import_paths()

    # Restore temporarily deleted task
    for k, v in tasks_to_hide.items():
        task_registry._tasks[k] = v



import json
import os
import re
import zipfile
from typing import Any, Dict
from unittest.mock import Mock, patch

import boto3
import moto
from moto.core.models import base_decorator
from moto.glue.responses import GlueResponse

from redun import File, task
from redun.config import Config
from redun.executors.aws_glue import (
    AWSGlueExecutor,
    get_redun_lib_files,
    package_redun_lib,
)
from redun.executors.aws_utils import get_aws_client
from redun.executors.code_packaging import create_zip
from redun.scheduler import Job
from redun.tests.utils import mock_s3, mock_scheduler, use_tempdir, wait_until
from redun.utils import pickle_dumps


class RedunGlueBackend(moto.glue.models.GlueBackend):
    """
    Moto doesn't implement all the glue methods we need, so this class,
    combined with RedunGlueResponse, adds the necessary mock functionality.
    Submitted jobs automatically start in the RUNNING state and don't change
    state, but it's enough to test the state of the queue for our purposes.

    For strange moto internal reasons this is NOT thread safe and the internal
    glue state depends on which thread you're calling from :(
    """

    from moto.core import BaseBackend, BaseModel

    def __init__(self) -> None:
        self.job_id_num = 0
        self.job_defs: Dict[str, dict] = {}  # Job definitions.
        super(RedunGlueBackend, self).__init__()

    @property
    def client(self):
        return get_aws_client("glue")

    @property
    def urls(self):
        urls = {}
        for url_base in self.url_bases:
            for url_path, handler in self.url_paths.items():
                url = url_path.format(url_base)
                urls[url] = handler
        return urls

    @property
    def url_bases(self):
        return ["https?://glue.(.+).amazonaws.com"]

    @property
    def url_paths(self):
        return {"{0}/$": RedunGlueResponse.dispatch}

    def get_job(self, JobName: str) -> Dict[str, Any]:
        if JobName in self.job_defs:
            return {"Job": self.job_defs[JobName]}

        raise self.client.exceptions.EntityNotFoundException(
            error_response={"Error": {"Code": "blah", "Message": "oops", "client": self.client}},
            operation_name="get_job",
        )

    def get_job_run(self, **kwargs):
        return {
            "JobRun": {
                # "JobRunState": "SUCCEEDED",
                "JobRunState": "RUNNING",
                "Id": kwargs["RunId"],
            }
        }

    def create_job(self, **kwargs) -> Dict[str, str]:
        assert "Name" in kwargs
        jobname = kwargs["Name"]

        self.job_defs[jobname] = kwargs

        return {"Name": jobname}

    def start_job_run(
        self,
        JobName: str,
        Arguments: Dict,
        Timeout: int,
        WorkerType: str,
        NumberOfWorkers: int,
        **kwargs,
    ) -> Dict[str, str]:
        if JobName not in self.job_defs:
            raise self.client.exceptions.EntityNotFoundException(
                {
                    "Error": {
                        "Code": "ENOENT",
                        "Message": (
                            f"{JobName} not in definitions " f"{self.job_defs.keys()} for {self}"
                        ),
                    }
                },
                "start_job_run",
            )

        self.job_id_num += 1
        return {"JobRunId": str(self.job_id_num)}


redun_glue_backend = RedunGlueBackend()
mock_redun_glue = base_decorator({"global": redun_glue_backend})


class RedunGlueResponse(GlueResponse):
    @property
    def glue_backend(self):
        return redun_glue_backend

    def get_job(self):
        name = self.parameters.get("JobName")
        job = self.glue_backend.get_job(name)
        return json.dumps(job)

    def create_job(self):
        return json.dumps(self.glue_backend.create_job(**self.parameters))

    def start_job_run(self):
        return json.dumps(self.glue_backend.start_job_run(**self.parameters))

    def get_job_run(self):
        return json.dumps(self.glue_backend.get_job_run(**self.parameters))


# Creates a fresh redun glue backend for each test.
def setup_function():
    redun_glue_backend.jobs = {}


@moto.mock_s3
@mock_redun_glue
def mock_executor(scheduler, debug=False, code_package=False):
    """
    Returns an AWSGlueExecutor with AWS API mocks.
    """

    s3_scratch_prefix = "s3://example-bucket/redun/"
    config = Config(
        {
            "glue": {
                "s3_scratch": s3_scratch_prefix,
                "role": "arn:aws:iam::123:role/service-role/AWSGlueServiceRole",
                "job_monitor_interval": 1.0,
                "job_retry_interval": 0.5,
                "code_package": code_package,
                "debug": debug,
                "glue_job_name": "test_glue_job",
                "glue_code_location": "s3://example-bucket/",
                "redun_zip_path": "s3://example-bucket/redun.zip",
            }
        }
    )
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    with File("s3://example-bucket/redun.zip").open("w") as fh:
        fh.write("I am definitely a real file")

    executor = AWSGlueExecutor("glue", scheduler, config["glue"])

    executor.get_jobs = Mock()
    executor.get_jobs.return_value = []

    return executor


@task()
def task1(x: int):
    return x + 10


@mock_s3
@mock_redun_glue
def test_executor_config(scheduler):
    config = Config(
        {
            "glue": {
                "s3_scratch": "s3_scratch_prefix",
                "aws_region": "pangea",
                "role": "overlord",
                "job_monitor_interval": 1.0,
                "code_includes": "*.txt",
                "glue_job_name": "test_glue_job",
                "glue_code_location": "s3://example-bucket/",
                "redun_zip_path": "s3://example-bucket/redun.zip",
                "glue_version": "3.0",
            }
        }
    )
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    with File("s3://example-bucket/redun.zip").open("w") as fh:
        fh.write("I am definitely a real file")

    executor = AWSGlueExecutor("glue", scheduler, config["glue"])

    assert executor.aws_region == "pangea"
    assert executor.role == "overlord"
    assert executor.interval == 1.0
    assert executor.code_package["includes"] == ["*.txt"]
    assert executor.debug is False
    assert executor.glue_version == "3.0"


@use_tempdir
def test_redun_lib_files() -> None:
    """
    We should be able to create a zip archive of the redun library.
    """
    lib_files = list(get_redun_lib_files())
    assert len(lib_files) > 0

    base_path = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    zip_file = create_zip("redun.zip", base_path, lib_files)
    zip_obj = zipfile.ZipFile(zip_file.path)
    namelist = zip_obj.namelist()
    assert len(namelist) == len(lib_files)
    assert "redun/__init__.py" in namelist


@mock_s3
def test_package_redun_lib() -> None:
    """
    We should be able to package the redun lib to S3.
    """
    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")
    s3_scratch_prefix = "s3://example-bucket/redun/"

    redun_file = package_redun_lib(s3_scratch_prefix)
    assert re.match("s3://example-bucket/redun/glue/redun-[0-9a-f]{40}.zip", redun_file.path)


@mock_s3
@mock_redun_glue
@patch("redun.executors.aws_glue.glue_describe_jobs")
@patch("redun.executors.aws_glue.submit_glue_job")
def test_executor_inflight_glue_job(submit_job_mock, describe_jobs_mock) -> None:
    """
    Ensure we reunite with an inflight glue job
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    assert executor.preexisting_glue_jobs == {}

    # Set up mocks to indicate a job is inflight.
    executor.get_jobs.return_value = [
        {
            "Id": "carrots",
            "Arguments": {"--job-hash": "hashbrowns"},
            "JobRunState": "RUNNING",
        }
    ]
    describe_jobs_mock.return_value = iter(
        [
            {
                "Id": "carrots",
                "Arguments": {"--job-hash": "hashbrowns"},
                "JobRunState": "RUNNING",
            }
        ]
    )

    # Force scheduler to gather inflight jobs.
    executor.gather_inflight_jobs()

    assert executor.preexisting_glue_jobs == {"hashbrowns": "carrots"}

    # Submit a job with same hash and ensure it isn't actually submitted
    rjob = Job(task1, task1(10))
    rjob.id = "123"
    rjob.eval_hash = "hashbrowns"
    rjob.args = ((10,), {})
    executor.submit(rjob)
    executor.stop()
    executor._monitor_thread.join()
    executor._submit_thread.join()

    # Ensure no new job was submitted.
    assert submit_job_mock.call_count == 0
    assert executor.running_glue_jobs["carrots"] == rjob


@mock_s3
@mock_redun_glue
def test_job_def_creation() -> None:
    """
    Tests that the job definition and oneshot file are created
    on executor init.
    """

    # Code file should not exist.
    client = boto3.client("glue", region_name="us-east-1")
    s3client = boto3.client("s3", region_name="us-east-1")
    s3client.create_bucket(Bucket="example-bucket")

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)

    executor.get_or_create_job_definition()

    # Now job def should exist
    retval = client.get_job(JobName=executor.glue_job_name)
    assert retval["Job"]["Name"] == executor.glue_job_name
    assert retval["Job"]["Role"] == executor.role
    assert re.match(
        "s3://example-bucket/redun/glue/oneshot-.*.py",
        retval["Job"]["Command"]["ScriptLocation"],
    )

    # Glue oneshot should now exist on S3.
    codefile = File(retval["Job"]["Command"]["ScriptLocation"])
    assert codefile.exists()


@mock_redun_glue
@mock_s3
@patch("redun.executors.aws_glue.glue_describe_jobs")
def test_glue_submission_retry(describe_jobs_mock) -> None:
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    assert executor.preexisting_glue_jobs == {}

    # Make job submission return None and test resubmit works
    do_fail = True

    def get_retval(*args, **kwargs):
        if do_fail:
            return None
        return "potato"

    executor.submit_pending_job = Mock()
    executor.submit_pending_job.side_effect = get_retval

    # Submit a job and make sure it is in pending queue.
    rjob = Job(task1, task1(10))
    rjob.id = "123"
    rjob.eval_hash = "hashbrowns"
    rjob.args = ((10,), {})

    executor.submit(rjob)

    wait_until(lambda: executor.submit_pending_job.call_count > 0)
    assert list(executor.pending_glue_jobs) == [rjob]
    assert len(executor.running_glue_jobs) == 0

    # Now when job submission works, it should be removed from pending queue
    do_fail = False
    wait_until(
        lambda: (
            len(executor.pending_glue_jobs) == 0
            and list(executor.running_glue_jobs.keys()) == ["potato"]
        )
    )

    # Simulate AWS Glue completing with valid value.
    output_file = File("s3://example-bucket/redun/jobs/hashbrowns/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Update mocks to indicate a job is done.
    executor.get_jobs.return_value = [
        {
            "Id": "potato",
            "Arguments": {"--job-hash": "hashbrowns"},
            "JobRunState": "SUCCEEDED",
        }
    ]
    describe_jobs_mock.return_value = iter(
        [
            {
                "Id": "potato",
                "Arguments": {"--job-hash": "hashbrowns"},
                "JobRunState": "SUCCEEDED",
            }
        ]
    )

    scheduler.batch_wait([rjob.id])
    executor.stop()
    executor._monitor_thread.join()
    executor._submit_thread.join()

    assert not scheduler.job_errors.get(rjob.id)
    assert scheduler.job_results[rjob.id] == 20


@mock_redun_glue
@mock_s3
@patch("redun.executors.aws_glue.submit_glue_job")
def test_glue_submit_job(submit_job_mock) -> None:
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler, code_package=True)

    job = Job(task1, task1(10))
    job.id = "123"
    job.eval_hash = "eval_hash"
    job.args = ((10,), {})

    executor.submit(job)
    assert executor.is_running
    assert File("s3://example-bucket/redun/jobs/eval_hash/input").exists()

    executor.stop()
    executor._monitor_thread.join()
    executor._submit_thread.join()



from redun.context import get_context
from redun.executors.alias import AliasExecutor
from redun.executors.aws_batch import AWSBatchExecutor
from redun.executors.aws_glue import AWSGlueExecutor
from redun.executors.docker import DockerExecutor

try:
    from redun.executors.k8s import K8SExecutor
except (ImportError, ModuleNotFoundError):
    # Skip k8s executor if kubernetes is not installed.
    pass
try:
    from redun.executors.gcp_batch import GCPBatchExecutor
except (ImportError, ModuleNotFoundError):
    # Skip gcp_batch executor if google-cloud-batch is not installed.
    pass
from redun.executors.local import LocalExecutor
from redun.file import Dir, File, ShardedS3Dataset
from redun.handle import Handle
from redun.namespace import get_current_namespace, namespace
from redun.scheduler import (
    Scheduler,
    apply_tags,
    catch,
    cond,
    get_current_scheduler,
    merge_handles,
    set_current_scheduler,
    throw,
)
from redun.scripting import script
from redun.task import PartialTask, Task, get_task_registry, task
from redun.version import version

__version__ = version
__all__ = [
    "Dir",
    "File",
    "Handle",
    "PartialTask",
    "Scheduler",
    "ShardedS3Dataset",
    "Task",
    "apply_tags",
    "catch",
    "cond",
    "get_current_namespace",
    "get_current_scheduler",
    "get_task_registry",
    "merge_handles",
    "namespace",
    "script",
    "set_current_scheduler",
    "task",
    "throw",
    "version",
]



"""
Tests to confirm that redun.file backends support forking subprocesses, not just spawn.
"""

import multiprocessing as mp
import os
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from io import BytesIO
from typing import List, Literal
from unittest.mock import MagicMock, Mock, patch

import adlfs
import boto3
import fsspec.asyn
import pytest
from tqdm import trange

import redun
from redun import File
from redun.tests.utils import mock_s3

NOT_SET = "NOT SET"
S3_PATH = os.environ.get("REDUN_TEST_FILE_FORK_S3_PATH", NOT_SET)
AZ_PATH = os.environ.get("REDUN_TEST_FILE_FORK_AZ_PATH", NOT_SET)


def mock_azure_blob_file_system():
    m = MagicMock(adlfs.AzureBlobFileSystem)
    m.return_value.open = lambda *args, **kwargs: BytesIO(b"hello")
    return m


def reset_redun_file():
    redun.file._filesystem_instances.clear()


@pytest.mark.parametrize(
    "path, multiprocessing_context",
    [(S3_PATH, "fork"), (S3_PATH, "spawn"), (AZ_PATH, "spawn"), (AZ_PATH, "fork")],
)
@patch("fsspec.asyn.reset_lock", side_effect=fsspec.asyn.reset_lock)
def test_supports_fork_integration(
    reset_lock: Mock, path: str, multiprocessing_context: Literal["spawn", "fork"]
):
    if path is NOT_SET:
        pytest.skip(
            "Please set environment variables REDUN_TEST_FILE_FORK_S3_PATH "
            "and REDUN_TEST_FILE_FORK_AZ_PATH to run this test."
        )

    reset_redun_file()

    num_processes = 2
    assert File(path).exists()

    actual = run_dataloaders(
        path, multiprocessing_context, num_processes=num_processes, num_items=2
    )
    assert actual == list(range(num_processes))


@mock_s3
@patch("fsspec.asyn.reset_lock", side_effect=fsspec.asyn.reset_lock)
def test_s3_supports_fork(reset_lock: Mock):
    reset_redun_file()

    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")
    path = "s3://example-bucket/mock.txt"
    file = File(path)
    file.write("hello")

    num_processes = 2
    assert File(path).exists()

    actual = run_dataloaders(path, "fork", num_processes=num_processes, num_items=2)
    assert actual == list(range(num_processes))


@pytest.mark.parametrize("multiprocessing_context", ["fork", "spawn", "none"])
@patch("adlfs.AzureBlobFileSystem")
@patch("fsspec.asyn.reset_lock", side_effect=fsspec.asyn.reset_lock)
@patch("redun.azure_utils.get_az_credential", return_value="mock_credential")
def test_az_supports_fork(
    mock_credential: Mock,
    reset_lock: Mock,
    azure_file_system: Mock,
    multiprocessing_context: Literal["fork", "spawn", "none"],
):
    reset_redun_file()

    azure_file_system.return_value.open = lambda *args, **kwargs: BytesIO(b"hello")

    path = "az://example-container@example-account.blob.core.windows.net/mock.txt"
    num_processes = 2
    assert File(path).exists()

    actual = run_dataloaders(
        path, multiprocessing_context, num_processes=num_processes, num_items=2
    )
    assert actual == (list(range(num_processes)) if multiprocessing_context != "none" else [-1])

    # The following assert tests that reset_lock is called at most once
    # If it's called multiple times, it means something's wrong. File access will work but be
    # slower than it should be. In reality, it's always going to be called at least once,
    # but on the CI/CD test run, it sometimes gets called on just the real method and
    # not the mock, so we miss it. If it actually never gets called, we'll know, because the
    # calls to open the file will hang!
    assert reset_lock.call_count <= 1

    azure_file_system.assert_called_once()


def run_single_dataloader(num_items: int, path: str, process_idx: int):
    if "mock" in path:
        return run_single_dataloader_mocked(num_items, path, process_idx)

    for item in trange(num_items):
        assert len(File(path).read("rb")) > 0

    return process_idx


@patch("adlfs.AzureBlobFileSystem")
@patch("fsspec.asyn.reset_lock", side_effect=fsspec.asyn.reset_lock)
@patch("redun.azure_utils.get_az_credential", return_value="mock_credential")
def run_single_dataloader_mocked(
    num_items: int,
    path: str,
    process_idx: int,
    mock_credential: Mock,
    reset_lock: Mock,
    azure_file_system: Mock,
):
    azure_file_system.return_value.open = lambda *args, **kwargs: BytesIO(b"hello")
    for item in trange(num_items):
        assert len(File(path).read("rb")) > 0
    if process_idx != -1:
        # See comments on the other reset_lock assert in test_az_supports_fork
        assert reset_lock.call_count <= 1
        if path.startswith("az://"):
            azure_file_system.call_count <= 1
    return process_idx


def run_dataloaders(
    path: str,
    multiprocessing_context: Literal["spawn", "fork", "none"],
    num_processes: int = 2,
    num_items: int = 100,
) -> List[int]:
    assert File(path).exists()

    if num_processes == 0 or multiprocessing_context == "none":
        return [run_single_dataloader(num_items, path, -1)]

    mp_context = dict(spawn=mp.context.SpawnContext, fork=mp.context.ForkContext)[
        multiprocessing_context
    ]()

    with ProcessPoolExecutor(max_workers=num_processes, mp_context=mp_context) as pool:
        fcn = partial(run_single_dataloader, num_items, path)
        results = list(pool.map(fcn, range(num_processes)))

    return results



import json
import os
from typing import Any, Dict, Optional, Tuple

import requests

from redun import File, Scheduler
from redun.config import Config
from redun.executors.aws_utils import DEFAULT_AWS_REGION, get_aws_client
from redun.executors.base import get_executor_from_config
from redun.executors.launch import launch_script
from redun.executors.scratch import SCRATCH_INPUT, get_execution_scratch_file
from redun.expression import SchedulerExpression, TaskExpression
from redun.promise import Promise
from redun.scheduler import Execution, Job, subrun
from redun.task import CacheScope, scheduler_task
from redun.utils import pickle_dump, str2bool


class InvocationException(Exception):
    """
    Custom exception that is raised when there is an error encountered invoking an AWS lambda.
    """

    pass


@scheduler_task(namespace="redun")
def federated_task(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    entrypoint: str,
    *task_args,
    **task_kwargs,
) -> Promise:
    """Execute a task that has been indirectly specified in the scheduler config file by providing
    a `federated_task` and its executor. This allows us to invoke code we do not have locally.

    Since the code isn't visible, the cache_scope for the subrun is always set to CSE.

    Parameters
    ----------
    entrypoint: str
        The name of the `federated_task` section in the config that identifies the task to perform.
    task_args: Optional[List[Any]]
        Positional arguments for the task
    task_kwargs: Any
        Keyword arguments for the task
    """
    federated_task_configs = scheduler.config.get("federated_tasks")
    entrypoint_config = _get_federated_config(federated_task_configs, entrypoint)

    # This argument needs to be parsed.
    if "new_execution" in entrypoint_config:
        entrypoint_config["new_execution"] = str2bool(entrypoint_config["new_execution"])

    # Since we require a namespace, we can simply dot format this ourselves.
    wrapper_task_expr: TaskExpression = TaskExpression(
        task_name=f'{entrypoint_config.pop("namespace")}.{entrypoint_config.pop("task_name")}',
        args=task_args,
        kwargs=task_kwargs,
    )

    subrun_task = subrun(
        expr=wrapper_task_expr,
        executor=entrypoint_config.pop("executor"),
        load_modules=[entrypoint_config.pop("load_module")],
        # Since the code isn't visible, only accept CSE hits.
        cache_scope=CacheScope.CSE,
        **entrypoint_config,
    )

    return scheduler.evaluate(subrun_task, parent_job=parent_job)


@scheduler_task(namespace="redun")
def lambda_federated_task(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    config_name: str,
    entrypoint: str,
    lambda_function_name: str,
    scratch_prefix: str,
    dryrun: bool,
    *task_args,
    **task_kwargs,
) -> Promise[Tuple[str, Dict]]:
    """Submit a task to an AWS Lambda Function, by packaging the inputs and sending as the payload
    with instructions.

    This allows a workflow to trigger execution of a remote workflow, to be executed by another
    service.

    This is fire-and-forget, you do not get back the results. We do not implement a way to wait
    for results, because it's not currently possible to monitor the backend database and gracefully
    determine whether an execution has finished. Specifically, executions look "errored"
    until they succeed, so we would have to rely on timeouts to determine if an error actually
    happened. Or, we would need to create another side-channel to record progress, which we
    did not want to undertake.

    Parameters
    ----------
    config_name : str
        The name of the config to use; the lambda defines how this name is chosen, not redun.
    entrypoint : str
        The name of the entrypoint into the above config.
    url : str
        The name of the AWS Lambda function to invoke
    scratch_prefix : str
        The path where the packaged inputs can be written. This is dictated by the lambda, hence
        is a user input.
    dryrun : bool
        If false, actually invoke the lambda. If true, skip it.

    Other args are intended for the task itself.

    Returns
    -------
        str
            The execution ID we asked the proxy lambda to use
        Dict
            The data package.
    """
    promise: Promise[Tuple[str, Dict]] = Promise()

    execution = Execution()

    input_path = get_execution_scratch_file(scratch_prefix, execution.id, SCRATCH_INPUT)
    with File(input_path).open("wb") as out:
        pickle_dump([task_args, task_kwargs], out)

    request_data = {
        "config_name": config_name,
        "entrypoint": entrypoint,
        "input_path": input_path,
        "execution_id": execution.id,
    }

    if not dryrun:
        aws_region = os.environ.get("AWS_REGION", DEFAULT_AWS_REGION)
        lambda_client = get_aws_client("lambda", aws_region)
        response = lambda_client.invoke(
            FunctionName=lambda_function_name,
            Payload=json.dumps(request_data),
        )
        if not response["StatusCode"] == 200:
            raise InvocationException(
                f"Failed to invoke lambda '{lambda_function_name}'. Encountered erorr "
                f"{response['FunctionError']}: {response['Payload']}"
            )

    promise.do_resolve((execution.id, request_data))
    return promise


@scheduler_task(namespace="redun")
def rest_federated_task(
    scheduler: Scheduler,
    parent_job: Job,
    sexpr: SchedulerExpression,
    config_name: str,
    entrypoint: str,
    url: str,
    scratch_prefix: str,
    dryrun: bool,
    *task_args,
    **task_kwargs,
) -> Promise[Tuple[str, Dict]]:
    """Submit a task to a server over REST, by packaging the inputs and POSTing a data blob
    with instructions.

    This allows a workflow to trigger execution of a remote workflow, to be executed by another
    service.

    This is fire-and-forget, you do not get back the results. We do not implement a way to wait
    for results, because it's not currently possible to monitor the backend database and gracefully
    determine whether an execution has finished. Specifically, executions look "errored"
    until they succeed, so we would have to rely on timeouts to determine if an error actually
    happened. Or, we would need to create another side-channel to record progress, which we
    did not want to undertake.

    Parameters
    ----------
    config_name : str
        The name of the config to use; the server defines how this name is chosen, not redun.
    entrypoint : str
        The name of the entrypoint into the above config.
    url : str
        The URL to POST the result to.
    scratch_prefix : str
        The path where the packaged inputs can be written. This is dictated by the server, hence
        is a user input.
    dryrun : bool
        If false, actually perform the POST. If true, skip it.

    Other args are intended for the task itself.

    Returns
    -------
        str
            The execution ID we asked the proxy to use
        Dict
            The data package.
    """
    promise: Promise[Tuple[str, Dict]] = Promise()

    execution = Execution()

    input_path = get_execution_scratch_file(scratch_prefix, execution.id, SCRATCH_INPUT)
    with File(input_path).open("wb") as out:
        pickle_dump([task_args, task_kwargs], out)

    request_data = {
        "config_name": config_name,
        "entrypoint": entrypoint,
        "input_path": input_path,
        "execution_id": execution.id,
    }

    if not dryrun:
        response = requests.post(url, data=request_data)
        if not response.ok:
            response.raise_for_status()

    promise.do_resolve((execution.id, request_data))
    return promise


def launch_federated_task(
    federated_config_path: str,
    entrypoint: str,
    task_args: Optional[Tuple] = None,
    task_kwargs: Optional[Dict] = None,
    input_path: Optional[str] = None,
    execution_id: Optional[str] = None,
) -> str:
    """
    Launch the described federated task. Among other purposes, this is designed to make it easy
    to implement a REST server that can process the messages from `rest_federated_task`.

    As with the `launch` CLI verb, we briefly start a scheduler to help with sending off the work,
    but immediately shut it down.

    Inputs may be provided in explicit form via `task_args` and `task_kwargs`, or pre-packaged,
    but not both.

    Parameters
    ----------
    federated_config_path : str
        The path to the federated config file to use.
    entrypoint : str
        The name of the entrypoint into the above config.
    task_args : Optional[Tuple]
        Arguments to package and pass to the task.
    task_kwargs : Optional[Dict]
        Arguments to package and pass to the task.
    input_path : Optional[str]
        The path to already-packaged inputs, which must be suitable for use with the
        `redun launch` flag `--input`. This allows a server to forward inputs created by another
        application. This file is not opened, because we don't assume we have the code to
        unpickle the objects inside.
    execution_id : Optional[str]
        If provided, use this execution id. Otherwise, one will be generated for you.

    Returns
    -------
        str
            The execution id we asked the redun invocation to use.
    """
    execution = Execution(execution_id)

    config = Config()
    config.read_path(os.path.join(federated_config_path, "redun.ini"))

    federated_task_configs = config.get("federated_tasks")
    entrypoint_config = _get_federated_config(federated_task_configs, entrypoint)

    executor_name = entrypoint_config["executor"]
    task_name = f"{entrypoint_config['namespace']}.{entrypoint_config['task_name']}"

    executor = get_executor_from_config(config.get("executors", {}), executor_name)
    if not input_path:
        assert (
            task_args is not None and task_kwargs is not None
        ), "Must provide actual data - set () and {} if no arguments are desired."
        assert isinstance(task_args, tuple), f"Expected a tuple but got: {task_args}"
        assert isinstance(task_kwargs, dict), f"Expected a dict but got: {task_kwargs}"
        input_path = get_execution_scratch_file(
            executor.scratch_root(), execution.id, SCRATCH_INPUT
        )
        with File(input_path).open("wb") as out:
            pickle_dump([task_args, task_kwargs], out)

    script_command = [
        "redun",
        "-c",
        entrypoint_config["config_dir"],
        "run",
        "--input",
        input_path,
        "--execution-id",
        execution.id,
        entrypoint_config["load_module"],
        task_name,
    ]

    launch_script(
        config,
        executor=executor,
        script_command=script_command,
        execution_id=execution.id,
    )

    return execution.id


def _get_federated_config(config: Config, entrypoint: str) -> Dict[str, Any]:
    """
    Helper. Extract the named federated config from the provided object. Raise an error
    if not found.
    """

    # Load the federated config and find the specified entrypoint within it.
    assert entrypoint in config, (  # type: ignore
        f"Could not find the entrypoint `{entrypoint}` "
        f"in the provided federated tasks. Found `{list(config.keys())}`"
    )

    entrypoint_config: Dict[str, Any] = dict(config[entrypoint])

    required_keys = {"namespace", "task_name", "load_module", "executor", "config_dir"}
    available_keys = set(entrypoint_config.keys())
    assert required_keys.issubset(available_keys), (
        f"Federated task entry `{entrypoint}` does not have the required keys, missing "
        f"`{required_keys.difference(available_keys)}`"
    )

    # The description is optional, but it shouldn't be in the hash. Just hide it from the
    # downstream subrun, so we don't have to deal with merging a user's `config_args`.
    entrypoint_config.pop("description", None)

    return entrypoint_config



import pickle
import sys
from collections import namedtuple
from typing import Any, Callable, NamedTuple, Tuple

import pytest

import redun
from redun import Task, task
from redun.expression import Expression
from redun.namespace import compute_namespace
from redun.scheduler import Scheduler, set_arg_defaults
from redun.task import (
    CacheCheckValid,
    CacheScope,
    Func,
    get_task_registry,
    get_tuple_type_length,
    wraps_task,
)
from redun.tests.task_test_helpers import square_task
from redun.utils import get_func_source, pickle_dumps
from redun.value import get_type_registry


def test_task_config() -> None:
    """
    Task should be properly configured by decorator.
    """

    @task()
    def my_task():
        return 10

    assert compute_namespace(my_task) == my_task.namespace
    assert my_task.fullname == Task._format_fullname(compute_namespace(my_task), "my_task")

    @task(name="custom_name", namespace="my_namespace", load_module="custom.module")
    def my_task2():
        return 10

    assert compute_namespace(my_task2, namespace="my_namespace") == my_task2.namespace
    assert my_task2.load_module == "custom.module"

    redun.namespace("my_namespace2")

    @task()
    def my_task3():
        return 10

    assert compute_namespace(my_task3) == my_task3.namespace

    redun.namespace()

    assert my_task.name == "my_task"
    assert my_task.fullname == "my_task"
    assert my_task.load_module == "redun.tests.test_tasks"
    assert my_task.hash
    assert callable(my_task.func)
    assert (
        my_task.source
        == """\
    def my_task():
        return 10
"""
    )

    registry = redun.get_task_registry()
    assert registry.get(my_task.name) == my_task

    assert my_task2.name == "custom_name"
    assert my_task2.fullname == "my_namespace.custom_name"

    assert my_task3.fullname == "my_namespace2.my_task3"


def helper():
    print("hello")


def test_task_hashing():  # typing: ignore
    """
    Test various forms of Task hashing.
    """

    @task()
    def task1():
        return 10

    assert task1.get_hash() == "ec0f851458d39044ae657dd250fa9526209ae6df"

    @task(name="other_name")
    def task1():
        return 10

    assert task1.get_hash() == "a285fea4f954a7f695cc0f390afef52034dd8615"

    @task(name="other_name", namespace="other_module")
    def task1():
        return 10

    assert task1.get_hash() == "4ef98ce3cc62f8dc741be21bc7a212b907c50a8e"

    @task(name="other_name", namespace="other_module", version="1")
    def task1():
        return 10

    assert task1.get_hash() == "c4a9e6d68a9ea97c2a7cb5dee6bff084ea2ffcfc"

    @task(name="other_name", namespace="other_module", version="1")
    def task1():
        return "totaly different implementation"

    assert task1.get_hash() == "c4a9e6d68a9ea97c2a7cb5dee6bff084ea2ffcfc"

    @task(name="another_name", namespace="another_module", version="1")
    def task1():
        return "totaly different implementation"

    assert task1.get_hash() == "48ce769e8b759be9812a19cfdb69898678eee0b3"

    # Hash includes are order invariant
    @task(hash_includes=["a", 2])
    def task1():
        return 10

    assert task1.get_hash() == "36fc0f089161ea4142ef20272e581ae8fc0b1921"

    @task(hash_includes=[2, "a"])
    def task1():
        return 10

    assert task1.get_hash() == "36fc0f089161ea4142ef20272e581ae8fc0b1921"

    # Testing that a change in helper source code will modify the task hash
    global helper

    @task(hash_includes=[helper])
    def task1():
        return 10

    assert task1.get_hash() == "3177281599f4961164534cd2f9d94a892db123b2"

    helper.__code__ = (lambda _: print("hello! i am different now!")).__code__

    @task(hash_includes=[helper])
    def task1():
        return 10

    assert task1.get_hash() == "a11d9a4abf033b0ced1f9e766f407b3c6151ed29"


def test_version() -> None:
    """
    Hash depends on version instead function source.
    """

    @task(name="task1", version="1")
    def task1():
        return "hello"

    @task(name="task1", version="1")
    def task2():
        return "bye"

    assert task1.hash == task2.hash


def test_set_default_args() -> None:
    @task()
    def myfunc(arg1, arg2=10):
        return None

    args, kwargs = set_arg_defaults(myfunc, (1,), {"arg2": 2})
    assert args == (1,)
    assert kwargs == {"arg2": 2}

    args, kwargs = set_arg_defaults(myfunc, (1,), {})
    assert args == (1,)
    assert kwargs == {"arg2": 10}

    args, kwargs = set_arg_defaults(myfunc, (), {"arg1": 1})
    assert args == ()
    assert kwargs == {"arg1": 1, "arg2": 10}


def test_cache_task() -> None:
    """
    Ensure we can cache a task.
    """

    @task()
    def task1():
        return 10

    assert task1.__getstate__() == {
        "name": "task1",
        "namespace": "",
        "version": None,
        "hash": task1.hash,
        "compat": [],
        "script": False,
        "task_options": {},
    }

    # Serialize and deserialize the task.
    data = pickle_dumps(task1)
    task1b = pickle.loads(data)

    # Ensure task deserialized correctly.
    assert isinstance(task1b, Task)
    assert task1b.fullname == task1.fullname
    assert task1b.func == task1.func
    assert task1b.source == task1.source

    # The task is still valid because we have not changed its code since serialization.
    assert task1b.is_valid()

    # Only change task_options.
    @task(memory=4)  # type: ignore
    def task1():
        return 10

    # Deserialize again.
    task1c = pickle.loads(data)
    assert isinstance(task1c, Task)
    assert task1c.fullname == task1.fullname
    assert task1c.func == task1.func

    # Task should still be valid.
    assert task1c.is_valid()

    # Task should have new task options.
    assert task1c.get_task_options() == {"memory": 4}

    # Change the definition of the task.
    @task()  # type: ignore
    def task1():
        return 20

    # Deserialize again.
    task1d = pickle.loads(data)
    assert isinstance(task1d, Task)
    assert task1d.fullname == task1.fullname
    assert task1d.func == task1.func

    # Since code changes, the task is not valid to use.
    assert not task1d.is_valid()


def test_cache_task_del() -> None:
    """
    Ensure we can deserialize a Task with deleted code.
    """
    state: dict = {
        "name": "test_cache_task_del_task",
        "namespace": None,
        "version": None,
        "hash": "HASH",
        "sync": False,
        "compat": [],
    }

    task1 = Task.__new__(Task)
    task1.__setstate__(state)

    # A Task that refers to non-existent code should not be valid.
    assert not task1.is_valid()

    # We should get an error calling the function.
    with pytest.raises(ValueError):
        task1.func()

    # Similarly, the corresponding Expression should not be valid either.
    expr = task1()
    assert isinstance(expr, Expression)
    assert not expr.is_valid()


def test_get_func_source() -> None:
    """
    get_func_source() should return the source of a function.
    """

    @task()
    def task1():
        return 10

    assert (
        get_func_source(task1.func)
        == """\
    def task1():
        return 10
"""
    )


def test_override_source() -> None:
    """
    Task source can be overridden
    """

    # A task's source can be overridden
    @task()
    def task1():
        return 10

    @task(source=get_func_source(task1.func))
    def task2():
        return 20

    assert task1.source == task2.source == get_func_source(task1.func)

    # A task's source is not pickled
    assert "source" not in task2.__getstate__()

    # But unpickling will reconstruct task.source from a same-name in-registry task.
    data = pickle_dumps(task2)
    task2b = pickle.loads(data)
    # Ensure task deserialized correctly.
    assert task2b.source == get_task_registry().get("task2").source


def test_hash_uses_source_instead_of_inspect() -> None:
    """For version=None tasks, the hash uses task.source instead of dynamic inspection."""

    # Use type ignore so we can redefine task3 multiple times
    @task
    def task3():
        return 30

    hash3a = task3.get_hash()

    @task  # type: ignore
    def task3():
        return 30

    hash3b = task3.get_hash()

    @task(source="some source")  # type: ignore
    def task3():
        return 30

    hash3c = task3.get_hash()

    assert hash3a == hash3b
    assert hash3a != hash3c


def test_task_options() -> None:
    """
    Task options should be available on the task and updates by options().
    """

    @task(executor="batch")
    def task1():
        return 10

    assert task1.has_task_option("executor")
    assert "batch" == task1.get_task_option("executor")
    assert "batch" == task1.get_task_option("executor", "default_executor")

    assert not task1.has_task_option("memory")
    assert task1.get_task_option("memory") is None
    assert 1000 == task1.get_task_option("memory", 1000)

    assert task1.get_task_options() == {"executor": "batch"}

    # The expression should not need any updated options.
    assert task1().task_expr_options == {}

    # When we apply an update to the task, it should appear on the new task and expression.
    task2 = task1.options(executor="local")

    assert task2.has_task_option("executor")
    assert "local" == task2.get_task_option("executor")
    assert "local" == task2.get_task_option("executor", "default_executor")

    assert not task2.has_task_option("memory")
    assert task2.get_task_option("memory") is None
    assert 1000 == task2.get_task_option("memory", 1000)

    assert task2.get_task_options() == {"executor": "local"}

    assert task2().task_expr_options == {"executor": "local"}

    # Additional updates should accumulate.
    task3 = task2.options(memory=4)

    assert task3.has_task_option("executor")
    assert "local" == task3.get_task_option("executor")
    assert "local" == task3.get_task_option("executor", "default_executor")

    assert task3.has_task_option("memory")
    assert 4 == task3.get_task_option("memory")
    assert 4 == task3.get_task_option("memory", 1000)

    assert task3.get_task_options() == {
        "executor": "local",
        "memory": 4,
    }

    assert task3().task_expr_options == {
        "executor": "local",
        "memory": 4,
    }


def test_task_options_hash() -> None:
    """
    _task_options_overrides should contribute to task hash.
    """

    @task()
    def task1(x):
        return x + 1

    task2 = task1.options(memory=4)
    assert task1.get_hash() != task2.get_hash()

    # When using versioned tasks, hash should update as well.

    @task(version="1")
    def task3(x):
        return x + 1

    task4 = task3.options(memory=4)
    assert task3.get_hash() != task4.get_hash()


def test_task_typing() -> None:
    """
    Tasks with typing should pass mypy.
    """

    @task()
    def task1(x: int) -> str:
        return str(x)

    @task()
    def task2(x: str) -> bool:
        return True

    expr = task1(10)
    task2(expr)


# Defining NameTuple at top-level scope to work around:
# https://github.com/python/mypy/issues/7281#issuecomment-763842951
class User2(NamedTuple):
    name: str
    age: int


def test_tuple_type_length() -> None:
    """
    get_tuple_type_length() should infer the length of a tuple type.
    """
    assert get_tuple_type_length(Tuple[int, int]) == 2
    assert get_tuple_type_length(Tuple[int, int, int]) == 3
    assert get_tuple_type_length(Tuple[()]) == 0
    assert get_tuple_type_length(Tuple) is None
    assert get_tuple_type_length(Tuple[Any, ...]) is None

    User = namedtuple("User", ["name", "age", "friend"])
    assert get_tuple_type_length(User) == 3

    assert get_tuple_type_length(User2) == 2


def test_naked_task() -> None:
    """
    A naked task decorator (no arguments) should also produce a Task.
    """

    @task
    def add(a: int, b: int) -> int:
        return a + b

    assert isinstance(add, Task)


def test_wraps_task() -> None:
    def doubled_task(offset: int = 0) -> Callable[[Func], Task[Func]]:
        @wraps_task(wrapper_hash_includes=[offset])
        def _doubled_task(inner_task: Task) -> Callable[[Task[Func]], Func]:
            def do_doubling(*task_args, **task_kwargs) -> Any:
                return (
                    inner_task.func(*task_args, **task_kwargs)
                    + inner_task.func(*task_args, **task_kwargs)
                    + offset
                )

            return do_doubling

        return _doubled_task

    # 1. Test that the task will get implicitly created
    @doubled_task()
    def value_task_unnested(x: int):
        return 1 + x

    # Check we can call it and arguments gets passed.
    assert 8 == value_task_unnested.func(3)
    assert 10 == value_task_unnested.func(4)

    # The tasks get registered.
    assert "value_task_unnested" in get_task_registry()._tasks
    assert "_doubled_task.value_task_unnested" in get_task_registry()._tasks

    # 2. Test that we can accept an explicit task
    @doubled_task()
    @task
    def value_task_nested(x: int):
        return 1 + x

    # Check we can call it and arguments gets passed.
    assert 8 == value_task_nested.func(3)
    assert 10 == value_task_nested.func(4)

    # The tasks get registered.
    assert "value_task_nested" in get_task_registry()._tasks
    assert "_doubled_task.value_task_nested" in get_task_registry()._tasks

    @doubled_task()
    def value_task(x: int):
        return 1 + x

    # 3. Demonstrate that we can return tasks lazily.
    def listify_task() -> Callable[[Func], Task[Func]]:
        @wraps_task(wrapper_name="renamed_task")
        def _listify_task(inner_task: Task) -> Callable[[Task[Func]], Func]:
            def make_list(*task_args, copies=1, **task_kwargs) -> Any:
                return [inner_task] * copies

            return make_list

        return _listify_task

    @listify_task()
    def float_task():
        return 3.1415

    assert "float_task" in get_task_registry()._tasks
    assert "renamed_task.float_task" in get_task_registry()._tasks

    float_task_inner = get_task_registry()._tasks["renamed_task.float_task"]
    assert float_task.func() == [float_task_inner]
    assert float_task.func(copies=3) == [
        float_task_inner,
        float_task_inner,
        float_task_inner,
    ]

    # 4. Try using a wrapper that is not defined in this module, so we can tell whether the
    # `load_module` points here, as it should. This is more likely representative of real use,
    # that the wrapper and its use will not be colocated.
    @square_task()
    @task(namespace="custom_namespace")
    def value_task2():
        return 7

    assert value_task2.func() == 49
    assert value_task2.load_module == sys.modules[__name__].__name__
    assert value_task2.namespace == "custom_namespace"

    assert "custom_namespace.value_task2" in get_task_registry()._tasks
    assert "custom_namespace.square_task.value_task2" in get_task_registry()._tasks

    # 5. Try deeper nesting with explicit task creating
    @listify_task()
    @doubled_task(offset=1)
    @task()
    def extra_nested():
        return -2

    extra_nested_middle_task = get_task_registry()._tasks["renamed_task.extra_nested"]
    assert extra_nested.func(copies=2) == [
        extra_nested_middle_task,
        extra_nested_middle_task,
    ]
    assert extra_nested_middle_task.func() == -3
    assert get_task_registry()._tasks["_doubled_task.renamed_task.extra_nested"].func() == -2

    # 6. Deeper nesting with implicit task creation
    @listify_task()
    @doubled_task(offset=1)
    def extra_nested_implicit():
        return -2

    extra_nested_middle_task = get_task_registry()._tasks["renamed_task.extra_nested_implicit"]
    assert extra_nested_implicit.func(copies=2) == [
        extra_nested_middle_task,
        extra_nested_middle_task,
    ]
    assert extra_nested_middle_task.func() == -3
    assert (
        get_task_registry()._tasks["_doubled_task.renamed_task.extra_nested_implicit"].func() == -2
    )

    # 7. Test hash propagation via source
    extra_nested_stored_hash = extra_nested.hash

    @listify_task()  # type: ignore
    @doubled_task(offset=1)
    def extra_nested():
        return -3

    # These are the same except at the inner-most level
    assert extra_nested_stored_hash != extra_nested.hash

    # 8. Test hash propagation via extra data
    @listify_task()  # type: ignore
    @doubled_task(offset=2)
    def extra_nested():
        return -2

    assert extra_nested_stored_hash != extra_nested.hash

    # 9. It's turtles all the way down...
    # These tasks have repeated names, which have helped catch several errors.
    @doubled_task()
    @doubled_task()
    @doubled_task()
    @doubled_task()
    @doubled_task()
    def turtles():
        return 2

    assert turtles.func() == 64

    # After all our renaming, make sure our registry is still a correct index of task names.
    for task_name, task_ in get_task_registry()._tasks.items():
        assert task_name == task_.fullname


def test_wraps_task_inner_task(scheduler: Scheduler) -> None:
    """
    wraps_task should maintain a reference to the inner task.
    """

    def doubled_task(offset: int = 0) -> Callable[[Func], Task[Func]]:
        @wraps_task(wrapper_hash_includes=[offset])
        def _doubled_task(inner_task: Task) -> Callable[[Task[Func]], Func]:
            def do_doubling(*task_args, **task_kwargs) -> Any:
                return (
                    inner_task.func(*task_args, **task_kwargs)
                    + inner_task.func(*task_args, **task_kwargs)
                    + offset
                )

            return do_doubling

        return _doubled_task

    @doubled_task()
    def plus_one(x: int):
        return 1 + x

    assert scheduler.run(plus_one(2)) == 6
    assert plus_one.hash == "eb330208b0723c0c478899869da03eb18d20ef33"

    assert plus_one.inner_task.hash == "16ebc98096f780d2fe24d0d6f56dfce29008f5a3"
    wrapped_task = plus_one.wrapped_task
    assert wrapped_task and wrapped_task.hash == "16ebc98096f780d2fe24d0d6f56dfce29008f5a3"
    assert plus_one.signature.parameters["x"].annotation == int

    @doubled_task()
    @doubled_task()
    def plus_two(x: int):
        return 2 + x

    assert scheduler.run(plus_two(2)) == 16

    assert plus_two.inner_task.hash == "5636d44845593eb5845f2ced8d204f1295d26653"
    assert plus_one.signature.parameters["x"].annotation == int


def test_use_wrapper_signature(scheduler: Scheduler) -> None:
    """
    Users can change the signature of the wrapped task.
    """

    @wraps_task(use_wrapper_signature=True)
    def doubled_task(inner_task: Task) -> Callable[[int, int], int]:
        def do_doubling(x: int, offset: int) -> Any:
            return inner_task.func(x) + inner_task.func(x) + offset

        return do_doubling

    @doubled_task
    def plus_one(x: int):
        return x + 1

    assert scheduler.run(plus_one(3, 1)) == 9
    assert plus_one.signature.parameters["x"].annotation == int
    assert plus_one.signature.parameters["offset"].annotation == int


def test_cache_options() -> None:
    """All the methods for setting cache options get sanitized into enums."""

    # Start with setting them on the method
    @task(cache=False)
    def task1():
        return None

    assert task1.get_task_option("cache_scope") == CacheScope.CSE

    @task(cache=True)
    def task2():
        return None

    assert task2.get_task_option("cache_scope") == CacheScope.BACKEND

    @task(cache_scope="CSE")
    def task3():
        return None

    assert task3.get_task_option("cache_scope") == CacheScope.CSE

    @task(cache_scope=CacheScope.CSE)
    def task4():
        return None

    assert task4.get_task_option("cache_scope") == CacheScope.CSE

    @task()
    def task5():
        return None

    # And retest with option overrides.
    assert task5.options(cache=False).get_task_option("cache_scope") == CacheScope.CSE
    assert task5.options(cache=True).get_task_option("cache_scope") == CacheScope.BACKEND
    assert task5.options(cache_scope="CSE").get_task_option("cache_scope") == CacheScope.CSE
    assert (
        task5.options(cache_scope=CacheScope.CSE).get_task_option("cache_scope") == CacheScope.CSE
    )
    with pytest.raises(ValueError, match="'typo' is not a valid CacheScope"):
        task5.options(cache_scope="typo")

    assert task5.options(check_valid="full").get_task_option("check_valid") == CacheCheckValid.FULL
    assert (
        task5.options(check_valid=CacheCheckValid.FULL).get_task_option("check_valid")
        == CacheCheckValid.FULL
    )
    assert (
        task5.options(check_valid="shallow").get_task_option("check_valid")
        == CacheCheckValid.SHALLOW
    )
    assert (
        task5.options(check_valid=CacheCheckValid.SHALLOW).get_task_option("check_valid")
        == CacheCheckValid.SHALLOW
    )
    with pytest.raises(ValueError, match="'typo' is not a valid CacheCheckValid"):
        task5.options(check_valid="typo")

    # Check the enum will traverse serialization.
    type_registry = get_type_registry()
    deserialized = type_registry.deserialize(
        "redun.Task",
        type_registry.serialize(task4.options(cache_scope=CacheScope.NONE)),
    )
    assert deserialized.get_task_option("cache_scope") == CacheScope.NONE



"""Make Job.execution_id non-nullable

Revision ID: cd2d53191748
Revises: d4af139b6f53
Create Date: 2021-08-01 07:18:29.549984

"""

import uuid
from typing import Any

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Column, ForeignKey, String, text
from sqlalchemy.orm import Session, declarative_base

# revision identifiers, used by Alembic.
revision = "cd2d53191748"
down_revision = "d4af139b6f53"
branch_labels = None
depends_on = None


Base: Any = declarative_base()


class Job(Base):
    __tablename__ = "job"

    id = Column(String, nullable=False, primary_key=True)


class Execution(Base):
    __tablename__ = "execution"

    id = Column(String, nullable=False, primary_key=True)
    args = Column(String, nullable=False)
    job_id = Column(String, ForeignKey("job.id", deferrable=True), nullable=False, index=True)


def upgrade():
    # Create stub executions for root jobs (parent_id is null) that have no execution.
    session = Session(bind=op.get_bind())
    try:
        job_ids = session.execute(
            text(
                """
            select job.id
            from job
            left join execution on execution.job_id = job.id
            where
              job.parent_id is null and
              execution.id is null
        """
            )
        ).fetchall()
        for (job_id,) in job_ids:
            session.add(Execution(id=str(uuid.uuid4()), args='"Stub Execution"', job_id=job_id))
        session.commit()
    finally:
        session.close()

    # Perform data migration to backfill Job.execution_id
    # using a recursive query through all parent_job pointers.
    op.execute(
        """
        -- Use a temp table to speed up the update.
        drop table if exists tmp_ancestors;
        """
    )

    op.execute(
        """
        create table tmp_ancestors as
        with recursive ancestors as (
            -- Get root jobs and their execution.
            select j.id as job_id, e.id as exec_id
            from job j
            join execution e on e.job_id = j.id
            union
            -- recurse.
            select j2.id, a.exec_id
            from job j2
            join ancestors a on a.job_id = j2.parent_id
        )
        select job_id, exec_id
        from ancestors;
        """
    )

    op.execute(
        """
        create index ix_tmp_ancestors_job_id on tmp_ancestors(job_id);
        """
    )

    op.execute(
        """
        update job
        set execution_id = (
            select a.exec_id
            from tmp_ancestors a
            where a.job_id = job.id
        );
        """
    )

    op.execute(
        """
        drop table tmp_ancestors;
        """
    )

    op.execute("commit")

    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("job", schema=None) as batch_op:
        batch_op.alter_column("execution_id", existing_type=sa.VARCHAR(), nullable=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("job", schema=None) as batch_op:
        batch_op.alter_column("execution_id", existing_type=sa.VARCHAR(), nullable=True)

    # ### end Alembic commands ###



import pytest

from redun import Scheduler, task
from redun.expression import SimpleExpression, TaskExpression, ValueExpression
from redun.functools import identity
from redun.task import scheduler_task


def test_value_expression(scheduler: Scheduler) -> None:
    """
    ValueExpression should evaluate to itself.
    """
    # Test evaluation.
    expr = ValueExpression(10)
    assert scheduler.run(expr) == 10

    # Test hashing.
    assert expr.get_hash() == "8ce63809f39a33d99358ccb2b18c8761d6764aea"

    # Test caching.
    state = expr.__getstate__()
    assert state == {
        "value_type": "builtins.int",
        "value": b"\x80\x03K\n.",
    }

    # Test deserialization.
    expr2 = ValueExpression.__new__(ValueExpression)
    expr2.__setstate__(state)
    assert expr2.value == 10

    # Test valid checking.
    assert expr.is_valid()


def test_simple_expression(scheduler: Scheduler) -> None:
    # Test evaluation.
    expr: SimpleExpression[str] = SimpleExpression("getitem", (["a", "b", "c"], 1))
    assert scheduler.run(expr) == "b"

    expr = SimpleExpression("getitem", (["a", "b", "c"], 5))
    with pytest.raises(IndexError):
        scheduler.run(expr)

    expr = SimpleExpression("bad_func", (10,))
    with pytest.raises(NotImplementedError):
        scheduler.run(expr)

    # Test hashing.
    expr = SimpleExpression("getitem", (["a", "b", "c"], 1))
    assert expr.get_hash() == "52d8a5c4b3672fc3dc7d53d5eb7a675a51f62894"

    # Test caching.
    state = expr.__getstate__()
    assert state == {
        "func_name": "getitem",
        "args": (
            b"\x80\x03]q\x00(X\x01\x00\x00\x00aq\x01X\x01\x00\x00\x00bq\x02X\x01"
            b"\x00\x00\x00cq\x03eK\x01\x86q\x04."
        ),
        "kwargs": b"\x80\x03}q\x00.",
    }

    expr2 = SimpleExpression.__new__(SimpleExpression)
    expr2.__setstate__(state)
    assert expr2.func_name == "getitem"
    assert expr2._upstreams == [(["a", "b", "c"], 1), {}]
    assert scheduler.run(expr2) == "b"

    # Test valid checking.
    assert expr2.is_valid()


class Inner:
    def __init__(self, a: int):
        self.a = a


def test_lazy_operators(scheduler: Scheduler) -> None:
    """
    Ensure expressions can be combined with operators into SimpleExpressions.
    """
    id = identity

    @task()
    def get_task():
        return identity

    expressions = [
        (id(0) == id(0), True),
        (id(1) != id(0), True),
        (id(2) < id(3), True),
        (id(2) <= id(3), True),
        (id(3) > id(2), True),
        (id(3) >= id(2), True),
        (id(1) + id(2), 3),
        (id(1) + 2, 3),
        (1 + id(2), 3),
        (id("a") + id("b"), "ab"),
        (id("a") + "b", "ab"),
        ("a" + id("b"), "ab"),
        (id(4) - id(3), 1),
        (id(4) - 3, 1),
        (4 - id(3), 1),
        (id(4) * id(3), 12),
        (id(4) * 3, 12),
        (4 * id(3), 12),
        (id(12) / id(3), 4),
        (id(12) / 3, 4),
        (12 / id(3), 4),
        (id(True) & id(False), False),
        (id(True) & id(True), True),
        (id(True) & True, True),
        (True & id(True), True),
        (id(False) | id(False), False),
        (id(True) | id(False), True),
        (id(True) | id(True), True),
        (id(True) | True, True),
        (True | id(True), True),
        (get_task()(7), 7),
        (id(Inner(7)).a, 7),
        (id(["a", "b", "c"])[1], "b"),
    ]

    for expr, expected in expressions:
        assert isinstance(expr, SimpleExpression)
        assert scheduler.run(expr) == expected


def test_task_expression(scheduler: Scheduler) -> None:
    @task()
    def task1(x, y=1):
        return x + y

    expr = task1(10)
    assert isinstance(expr, TaskExpression)
    assert expr.task_name == "task1"
    assert expr.args == (10,)
    assert expr.kwargs == {}
    assert scheduler.run(expr) == 11

    expr = task1(10, y=3)
    assert expr.kwargs == {"y": 3}
    assert scheduler.run(expr) == 13

    @task()
    def buggy():
        raise ValueError("bugs")

    with pytest.raises(ValueError):
        scheduler.run(buggy())

    # Test hashing.
    expr = TaskExpression("task1", (10,), {})
    assert expr.get_hash() == "7076d9536c59c05862286df0d767468098a4f0ed"

    # Test caching.
    state = expr.__getstate__()
    assert state == {
        "task_name": "task1",
        "args": b"\x80\x03K\n\x85q\x00.",
        "kwargs": b"\x80\x03}q\x00.",
        "task_options": {},
    }

    expr2 = TaskExpression.__new__(TaskExpression)
    expr2.__setstate__(state)
    assert expr2.task_name == "task1"
    assert scheduler.run(expr2) == 11

    # Test valid checking.
    assert expr2.is_valid()
    expr2.task_name = "bad_task"
    assert not expr2.is_valid()

    # Test call expressions.
    assert scheduler.run(SimpleExpression("call", (task1, (2,), {"y": 3}))) == 5


def test_ban_bool_coerce() -> None:
    """
    Coercision of an Expression to bool should be ban to avoid common bugs.
    """
    expr = ValueExpression(10)
    with pytest.raises(TypeError):
        if expr:
            pass


def test_ban_iter() -> None:
    """
    Ban iteration on expression to prevent common misuse.
    """

    @task()
    def task1():
        return [1, 2, 3]

    expr = task1()

    with pytest.raises(TypeError):
        for x in expr:
            pass

    with pytest.raises(TypeError):
        _ = iter(expr)


def test_repr() -> None:
    """
    Test the repr of an Expression matches the expected syntax.
    """

    @task
    def add(a, b):
        pass

    @scheduler_task()
    def cond(scheduler, parent_job, sexpr, pred, then, else_):
        pass

    # Common task calls.
    assert repr(add(1, 2)) == "add(1, 2)"
    assert repr(add(1, b=2)) == "add(1, b=2)"
    assert repr(add(a=True, b="2")) == "add(a=True, b='2')"
    assert repr(add([1, 2, 3], [4])) == "add([1, 2, 3], [4])"

    # Recursive task calls.
    assert repr(add(add(1, 2), add(3, 4))) == "add(add(1, 2), add(3, 4))"

    # SchedulerTask calls.
    assert repr(cond(True, add(1, 2), add(3, 4))) == "cond(True, add(1, 2), add(3, 4))"

    # Simple expression operators.
    assert repr(add(1, 2) + 3) == "(add(1, 2) + 3)"
    assert repr(3 + add(1, 2)) == "(3 + add(1, 2))"
    assert repr(add(1, 2) + 3 + 4) == "((add(1, 2) + 3) + 4)"
    assert repr(add(1, 2) == 3) == "(add(1, 2) == 3)"
    assert repr(add(1, 2) & add(3, 4)) == "(add(1, 2) & add(3, 4))"

    # Simple expressions: getitem, getattr, call.
    assert repr(add(1, 2)["key"]) == "add(1, 2)['key']"
    assert repr(add(1, 2)[1]) == "add(1, 2)[1]"
    assert repr(add(1, 2)[1:3]) == "[add(1, 2)[1], add(1, 2)[2]]"
    assert repr(add(1, 2)[:2]) == "[add(1, 2)[0], add(1, 2)[1]]"
    assert repr(add(1, 2).attr) == "add(1, 2).attr"
    assert repr(add(1, 2).func(1, 2)) == "add(1, 2).func(1, 2)"



import argparse

from redun.backends.db import Tag
from redun.console.parser import format_args, parse_args
from redun.console.utils import get_links


def test_format_args() -> None:
    """
    Ensure we can parse and format argv into argparse.Namespace and back.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--flag", action="store_true")
    parser.add_argument("--page", type=int, default=1)
    parser.add_argument("--file", action="append", default=[])

    argv = ["--flag", "--page", "3", "--file", "aa", "--file", "bb"]
    args = parse_args(parser, argv)
    assert args.flag
    assert args.page == 3
    assert args.file == ["aa", "bb"]
    assert format_args(parser, args) == argv

    argv = []
    args = parse_args(parser, argv)
    assert not args.flag
    assert args.page == 1
    assert args.file == []
    assert format_args(parser, args) == argv


def test_format_link() -> None:
    """
    Ensure tags can be used to generate links.
    """
    tags_dict = {
        "git_origin_url": "git@github.com:acme/workflow.git",
        "git_origin2_url": "https://github.com/acme/workflow.git",
        "git_commit": "96ab8a09853746fdc8b8b408a9de3647a92a9450",
        "aws_batch_job": "8d90f8fd-77fb-46d8-b099-bb31019f47da",
        "aws_log_stream": "workflow-redun-jd/default/ee83478a1d3643ce9ed6ad5b05b9c7a1",
    }
    tags = [Tag(key=key, value=value) for key, value in tags_dict.items()]

    link_patterns = [
        r"https://github.com/{{git_origin_url:.*github\.com:(.*)\.git$}}/commits/{{git_commit}}",
        r"https://github2.com/{{git_origin2_url:.*github\.com(:|/)(?P<val>.*)\.git$}}/commits/{{git_commit}}",  # noqa: E501
        r"https://console.aws.amazon.com/batch/home?#jobs/detail/{{aws_batch_job}}",
        r"https://console.aws.amazon.com/cloudwatch/home?#logEventViewer:group=/aws/batch/job;stream={{aws_log_stream}}",  # noqa: E501
        r"https://foo.com/{{foo}}",
        r"https://github.com/{{git_origin_url:.*bad-url.com:(.*)\.git$}}/commits/{{git_commit}}",
    ]

    assert get_links(link_patterns, tags) == [
        "https://github.com/acme/workflow/commits/96ab8a09853746fdc8b8b408a9de3647a92a9450",
        "https://github2.com/acme/workflow/commits/96ab8a09853746fdc8b8b408a9de3647a92a9450",
        "https://console.aws.amazon.com/batch/home?#jobs/detail/8d90f8fd-77fb-46d8-b099-bb31019f47da",  # noqa: E501
        "https://console.aws.amazon.com/cloudwatch/home?#logEventViewer:group=/aws/batch/job;stream=workflow-redun-jd/default/ee83478a1d3643ce9ed6ad5b05b9c7a1",  # noqa: E501
    ]
